import numexpr as ne
import healpy as hp
import numpy as np
import cosmotool as ct
import argparse
import h5py as h5
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import ksz
from ksz.constants import *
from cosmotool import interp3d
import numpy as np
from scipy import ndimage
from scipy.special import sinc


def move_direction_new(d_theta, d_phi, theta, phi):
  cos=np.cos
  sin=np.sin
  sqrt=np.sqrt

  amplitude = sqrt(d_theta*d_theta + d_phi*d_phi);
  cos_alpha = d_theta/amplitude;
  sin_alpha = d_phi/amplitude;

  if (amplitude == 0):
    return theta,phi

  cos_d = cos(amplitude);
  sin_d = sin(amplitude);

  cos_theta = cos(theta);
  sin_theta = sin(theta);

  cos_phi = cos(phi);
  sin_phi = sin(phi);

  basis = [
     [ cos_phi * sin_theta, sin_phi * sin_theta, cos_theta ],
     [ cos_phi * cos_theta, sin_phi * cos_theta, -sin_theta ],
     [ -sin_phi, cos_phi, 0 ]
   ]

  np0 = [ cos_d, cos_alpha*sin_d, sin_alpha*sin_d ]
  np1 = [ sum([basis[j][i] * np0[j] for j in xrange(3)]) for i in xrange(3) ]
  dnp = sqrt(sum([np1[i]**2 for i in xrange(3)]))

  theta = np.arccos(np1[2]/dnp);
  phi = np.arctan2(np1[1], np1[0]) % (2*np.pi);

  return theta,phi

def move_direction_new2(delta_theta, delta_phi, theta, phi):
	cos,sin,sqrt=np.cos,np.sin,np.sqrt
	
	grad_len = sqrt(delta_theta**2 + delta_phi**2)
	if grad_len==0:
		return theta,phi

	cth0 = cos(theta)
	sth0 = sin(theta)
	
	topbottom = 1 if (theta < 0.5*np.pi) else -1
	sinc_grad_len = sinc(grad_len)

	cth = topbottom*cos(grad_len) * cth0 - sinc_grad_len*sth0*delta_theta
	sth = max(1e-10, sqrt((1.0-cth)*(1.0+cth)) )

	phi = phi + np.arcsin(delta_phi * sinc_grad_len / sth)
	theta = np.arccos(cth)
	return theta,phi

move_direction = move_direction_new

def wrapper_impulsion(f):

  class _Wrapper(object):
    def __init__(self):
      pass
      
    def __getitem__(self,direction):
    
      if 'velocity' in f:
        return f['velocity'][:,:,:,direction]
      
      n = "p%d" % direction
      return f[n]

  return _Wrapper()


def build_unit_vectors(N):
  ii = np.arange(N,dtype=np.float64)/N - 0.5
  d = np.sqrt(ii[:,None,None]**2 + ii[None,:,None]**2 + ii[None,None,:]**2)
  d[N/2,N/2,N/2] = 1
  ux = ii[:,None,None] / d
  uy = ii[None,:,None] / d
  uz = ii[None,None,:] / d

  return ux,uy,uz

def compute_vcmb(l, b):
    # Motion is obtained from Tully (2007): sun_vs_LS + LS_vs_CMB
    motion = [-25.,-246.,277.];

    x = np.cos(l*np.pi/180) * np.cos(b*np.pi/180)
    y = np.sin(l*np.pi/180) * np.cos(b*np.pi/180)
    z = np.sin(b*np.pi/180)

    return x*motion[0] + y*motion[1] + z*motion[2]


def compute_vlg(l,b):

    motion = [-79,296,-36]; # [-86, 305, -33];

    x = np.cos(l*np.pi/180) * np.cos(b*np.pi/180)
    y = np.sin(l*np.pi/180) * np.cos(b*np.pi/180)
    z = np.sin(b*np.pi/180)

    return x*motion[0] + y*motion[1] + z*motion[2]


def generate_from_catalog(dmin,dmax,Nside,perturb=0.0,y=0.0,do_random=False,do_hubble=False,x=2.37,bright=-np.inf,bright_list=[],use_vlg=True,sculpt=-1):
  import progressbar as pbar
  
  cat = np.load("2m++.npy")

  cat['distance'] = cat['best_velcmb']
#  cat = cat[np.where((cat['distance']>100*dmin)*(cat['distance']<dmax*100))]

  deg2rad = np.pi/180
  Npix = 12*Nside**2
  xp,yp,zp = hp.pix2vec(Nside, np.arange(Npix))
  N2 = np.sqrt(xp**2+yp**2+zp**2)

  ksz_template = np.zeros(Npix, dtype=np.float64)
  ksz_mask = np.ones(Npix, dtype=np.uint8)
  if do_hubble:
    ksz_hubble_template = np.zeros(ksz_template.size, dtype=np.float64)

  for i in pbar.ProgressBar(maxval = cat.size, widgets=[pbar.Bar(), pbar.ETA()])(cat):
    # Skip too point sources
    if i['name'] in bright_list:
      print("Object %s is in bright list" % i['name'])
      continue

    if do_random:
      l = np.random.rand()*360
      b = np.arcsin(2*np.random.rand()-1)*180/np.pi
    else:
      l0,b0=i['gal_long'],i['gal_lat']

    l=ne.evaluate('l0*deg2rad')
    b=ne.evaluate('b0*deg2rad')

    dtheta,dphi = np.random.randn(2)*perturb
    theta,l=move_direction(dtheta,dphi,0.5*np.pi - b, l)
	
    b = 0.5*np.pi-theta

    x0 = np.cos(l)*np.cos(b)
    y0 = np.sin(l)*np.cos(b)
    z0 = np.sin(b)

    if use_vlg:
      vlg = i['best_velcmb'] - compute_vcmb(l0, b0) + compute_vlg(l0, b0)
      DA = vlg/100
    else:
      DA = i['best_velcmb'] / 100

    if DA < dmin or DA > dmax:
      continue

    Lgal = DA**2*10**(0.4*(tmpp_cat['Msun']-i['K2MRS']+25))

    M_K=i['K2MRS']-5*np.log10(DA)-25
    # Skip too bright galaxies
    if M_K < bright:
      continue

    profiler = ksz.KSZ_Isothermal(Lgal, x, y=y, sculpt=sculpt)

    idx0 = hp.query_disc(Nside, (x0,y0,z0), 3*profiler.rGalaxy/DA)

    xp1 = xp[idx0]
    yp1 = yp[idx0]
    zp1 = zp[idx0]
    N2_1 = N2[idx0]

    cos_theta = ne.evaluate('(x0*xp1+y0*yp1+z0*zp1)/(sqrt(x0**2+y0**2+z0**2)*(N2_1))')

    idx,idx_masked,m = profiler.projected_profile(cos_theta, DA)
    idx = idx0[idx]
    idx_masked = idx0[idx_masked]
    ksz_template[idx] += m
    ksz_mask[idx_masked] = 0
    if do_hubble:
      ksz_hubble_template[idx] += m*DA

  ne.evaluate('ksz_template*ksz_normalization', out=ksz_template)

  result =ksz_template, ksz_mask
  if do_hubble:
    ne.evaluate('ksz_hubble_template*ksz_normalization', out=ksz_hubble_template)
    return result + ( ksz_hubble_template,)
  else:
    return result

def get_args():
  parser=argparse.ArgumentParser(description="Generate Skymaps from CIC maps")
  parser.add_argument('--Nside', type=int, default=128)
  parser.add_argument('--minval', type=float, default=-0.5)
  parser.add_argument('--maxval', type=float, default=0.5)
  parser.add_argument('--depth_min', type=float, default=10)
  parser.add_argument('--depth_max', type=float, default=60)
  parser.add_argument('--ksz_map', type=str, required=True)
  parser.add_argument('--base_fig', type=str, default="kszfig.png")
  parser.add_argument('--build_dipole', action='store_true')
  parser.add_argument('--degrade', type=int, default=-1)
  parser.add_argument('--y',type=float,default=0.0)
  parser.add_argument('--x',type=float,default=2.37)
  parser.add_argument('--random', action='store_true')
  parser.add_argument('--perturb', type=float, default=0)
  parser.add_argument('--hubble_monopole', action='store_true')
  parser.add_argument('--remove_bright', type=float, default=-np.inf)
  parser.add_argument('--bright_file', type=str)
  parser.add_argument('--lg', action='store_true')
  parser.add_argument('--sculpt_beam', type=float, default=-1)
  return parser.parse_args()

def main():

  args = get_args()

  ff=plt.figure(1)
  plt.clf()
  v=[]

  print("Generating map...")

  with open("crap.txt", mode="r") as f:
    bright_list = [l.split('#')[0].strip(" \t\n\r") for l in f]

  if args.bright_file:
    with open(args.bright_file, mode="r") as f:
      idx_name = f.readline().split(',').index('name_2')
      bright_list = bright_list + [l.split(',')[idx_name] for l in f]

  print("Built bright point source list: " + repr(bright_list))
  r = generate_from_catalog(args.depth_min,args.depth_max,args.Nside,perturb=args.perturb,y=args.y,do_random=args.random,do_hubble=args.hubble_monopole,x=args.x,bright=args.remove_bright,use_vlg=args.lg,bright_list=bright_list,sculpt=args.sculpt_beam)
  hubble_map = None
  if args.hubble_monopole:
    proj,mask,hubble_map = r
  else:
    proj,mask = r

  if args.degrade > 0:
    proj *= mask
    proj = hp.ud_grade(proj, nside_out=args.degrade)
    if hubble_map is not None:
      hubble_map *= mask
      hubble_map = hp.ud_grade(hubble_map, nside_out=args.degrade)
    mask = hp.ud_grade(mask, nside_out=args.degrade)
    Nside = args.degrade
  else:
    Nside = args.Nside

  hp.write_map(args.ksz_map + ".fits", proj)
  hp.write_map(args.ksz_map + "_mask.fits", mask)

  if args.build_dipole:
    x,y,z=hp.pix2vec(Nside, np.arange(hp.nside2npix(Nside)))
    hp.write_map(args.ksz_map + "_x.fits", proj*x)
    hp.write_map(args.ksz_map + "_y.fits", proj*y)
    hp.write_map(args.ksz_map + "_z.fits", proj*z)

  if args.hubble_monopole:
    hp.write_map(args.ksz_map + "_hubble.fits", hubble_map)

  hp.mollview(proj*100*1e6, fig=1, coord='GG', cmap=plt.cm.coolwarm, title='', min=args.minval, 
              max=args.maxval)

  ff.savefig(args.base_fig)
  
main()



import healpy as hp
import numpy as np
import cosmotool as ct
import argparse
import h5py as h5
from matplotlib import pyplot as plt

Mpc=3.08e22
rhoc = 1.8864883524081933e-26  # m^(-3)
sigmaT = 6.6524e-29
mp = 1.6726e-27
lightspeed = 299792458.
v_unit = 1e3  # Unit of 1 km/s
T_cmb=2.725
h = 0.71
Y = 0.245  #The Helium abundance
Omega_matter = 0.26
Omega_baryon=0.0445

G=6.67e-11
MassSun=2e30
frac_electron = 1.0 # Hmmmm
frac_gas_galaxy = 0.14
mu = 1/(1-0.5*Y)

baryon_fraction = Omega_baryon / Omega_matter

ksz_normalization = T_cmb*sigmaT*v_unit/(lightspeed*mu*mp) * baryon_fraction
rho_mean_matter = Omega_matter * (3*(100e3/Mpc)**2/(8*np.pi*G)) 

parser=argparse.ArgumentParser(description="Generate Skymaps from CIC maps")
parser.add_argument('--boxsize', type=float, required=True)
parser.add_argument('--Nside', type=int, default=128)
parser.add_argument('--base_h5', type=str, required=True)
parser.add_argument('--base_fig', type=str, required=True)
parser.add_argument('--start', type=int, required=True)
parser.add_argument('--end', type=int, required=True)
parser.add_argument('--step', type=int, required=True)
parser.add_argument('--minval', type=float, default=-0.5)
parser.add_argument('--maxval', type=float, default=0.5)
parser.add_argument('--depth_min', type=float, default=10)
parser.add_argument('--depth_max', type=float, default=60)
parser.add_argument('--iid', type=int, default=0)
parser.add_argument('--ksz_map', type=str, required=True)
args = parser.parse_args()

L = args.boxsize
Nside = args.Nside

def build_sky_proj(density, dmax=60.,dmin=0,iid=0):

  N = density.shape[0]
  ix = (np.arange(N)-0.5)*L/N - 0.5 * L


  dist2 = (ix[:,None,None]**2 + ix[None,:,None]**2 + ix[None,None,:]**2)

  flux = density.transpose().astype(ct.DTYPE) # / dist2
  dmax=N*dmax/L
  dmin=N*dmin/L
  if iid == 0:
    shifter = np.array([0.5,0.5,0.5])
  else:
    shifter = np.array([0.,0.,0.])
    
  projsky1 = ct.spherical_projection(Nside, flux, dmin, dmax, integrator_id=iid, shifter=shifter)

  return projsky1*L/N

def build_unit_vectors(N):
  ii = np.arange(N)*L/N - 0.5*L
  d = np.sqrt(ii[:,None,None]**2 + ii[None,:,None]**2 + ii[None,None,:]**2)
  d[N/2,N/2,N/2] = 1
  ux = ii[:,None,None] / d
  uy = ii[None,:,None] / d
  uz = ii[None,None,:] / d

  return ux,uy,uz

for i in xrange(args.start,args.end,args.step):
  ff=plt.figure(1)
  plt.clf()
  with h5.File(args.base_h5 % i, mode="r") as f:
    p = f['velocity'][:]
    davg = np.average(np.average(np.average(f['density'][:],axis=0),axis=0),axis=0)
    p /= davg # Now we have momentum scaled to the mean density

  # Rescale by Omega_b / Omega_m
  p = p.astype(np.float64)
  print p.max(), p.min(), ksz_normalization, rho_mean_matter
  p *= -ksz_normalization*rho_mean_matter*1e6
  
  ux,uy,uz = build_unit_vectors(p.shape[0])
  pr = p[:,:,:,0] * ux + p[:,:,:,1] * uy + p[:,:,:,2] * uz 
  print p.max(), p.min()
  print pr.max()*Mpc, pr.min()*Mpc
  
  @ct.timeit_quiet
  def run_proj():
    return build_sky_proj(pr*Mpc, dmin=args.depth_min,dmax=args.depth_max,iid=args.iid)

  run_proj()

  hp.write_map(args.ksz_map % i, proj)

  hp.mollview(proj, fig=1, coord='CG', cmap=plt.cm.coolwarm, title='Sample %d' % i, min=args.minval, 
              max=args.maxval)

  ff.savefig(args.base_fig % i)



import numpy as np

def fourier_analysis(borg_vol):
  L = (borg_vol.ranges[1]-borg_vol.ranges[0])
  N = borg_vol.density.shape[0]
    
  return np.fft.rfftn(borg_vol.density)*(L/N)**3, L, N

def borg_upgrade_sampling(dhat, supersample):
    N = dhat.shape[0]
    N2 = N * supersample
    dhat_new = np.zeros((N2, N2, N2/2+1), dtype=np.complex128)

    hN = N/2
    dhat_new[:hN, :hN, :hN+1] = dhat[:hN, :hN, :]
    dhat_new[:hN, (N2-hN):N2, :hN+1] = dhat[:hN, hN:, :]
    dhat_new[(N2-hN):N2, (N2-hN):N2, :hN+1] = dhat[hN:, hN:, :]
    dhat_new[(N2-hN):N2, :hN, :hN+1] = dhat[hN:, :hN, :]

    return dhat_new, N2

def half_pixel_shift(borg, doshift=False):

    dhat,L,N = fourier_analysis(borg)
    if not doshift:
      return dhat, L

    return bare_half_pixel_shift(dhat, L, N)

def bare_half_pixel_shift(dhat, L, N, doshift=False):

#    dhat_new,N2 = borg_upgrade_sampling(dhat, 2)
#    d = (np.fft.irfftn(dhat_new)*(N2/L)**3)[1::2,1::2,1::2]
#    del dhat_new
#    dhat = np.fft.rfftn(d)*(L/N)**3
#    return dhat, L

#    dhat2 = np.zeros((N,N,N),dtype=np.complex128)
#    dhat2[:,:,:N/2+1] = dhat
#    dhat2[N:0:-1, N:0:-1, N:N/2:-1] = np.conj(dhat[1:,1:,1:N/2])
#    dhat2[0, N:0:-1, N:N/2:-1] = np.conj(dhat[0, 1:, 1:N/2])
#    dhat2[N:0:-1, 0, N:N/2:-1] = np.conj(dhat[1:, 0, 1:N/2])
#    dhat2[0,0,N:N/2:-1] = np.conj(dhat[0, 0, 1:N/2])

    ik = np.fft.fftfreq(N,d=L/N)*2*np.pi
    phi = 0.5*L/N*(ik[:,None,None]+ik[None,:,None]+ik[None,None,:(N/2+1)])
#    phi %= 2*np.pi
    phase = np.cos(phi)+1j*np.sin(phi)
    dhat = dhat*phase    
    dhat[N/2,:,:] = 0
    dhat[:,N/2,:] = 0
    dhat[:,:,N/2] = 0
    
    return dhat, L




import matplotlib
matplotlib.use('Agg')
import healpy as hp
import numpy as np
import cosmotool as ct
import argparse
import h5py as h5
from matplotlib import pyplot as plt

parser=argparse.ArgumentParser(description="Generate Skymaps from CIC maps")
parser.add_argument('--boxsize', type=float, required=True)
parser.add_argument('--Nside', type=int, default=128)
parser.add_argument('--base_cic', type=str, required=True)
parser.add_argument('--base_fig', type=str, required=True)
parser.add_argument('--start', type=int, required=True)
parser.add_argument('--end', type=int, required=True)
parser.add_argument('--step', type=int, required=True)
parser.add_argument('--minval', type=float, default=0)
parser.add_argument('--maxval', type=float, default=4)
parser.add_argument('--depth_min', type=float, default=10)
parser.add_argument('--depth_max', type=float, default=60)
parser.add_argument('--iid', type=int, default=0)
parser.add_argument('--proj_cat', type=bool, default=False)
args = parser.parse_args()

#INDATA="/nethome/lavaux/Copy/PlusSimulation/BORG/Input_Data/2m++.npy"
INDATA="2m++.npy"
tmpp = np.load(INDATA)

L = args.boxsize
Nside = args.Nside

def build_sky_proj(density, dmax=60.,dmin=0,iid=0):

  N = density.shape[0]
  ix = (np.arange(N)-0.5)*L/N - 0.5 * L

#  dist2 = (ix[:,None,None]**2 + ix[None,:,None]**2 + ix[None,None,:]**2)

  flux = density.transpose().astype(ct.DTYPE) # / dist2
  dmax=N*dmax/L
  dmin=N*dmin/L
  if iid == 0:
    shifter = np.array([0.5,0.5,0.5])
  else:
    shifter = np.array([0.,0.,0.])
    
  projsky1 = ct.spherical_projection(Nside, flux, dmin, dmax, integrator_id=iid, shifter=shifter)

  return projsky1*L/N

l,b = tmpp['gal_long'],tmpp['gal_lat']

l = np.radians(l)
b = np.pi/2 - np.radians(b)

dcmb = tmpp['velcmb']/100.

idx = np.where((dcmb>args.depth_min)*(dcmb<args.depth_max))


for i in xrange(args.start,args.end,args.step):
  ff=plt.figure(1)
  plt.clf()
  d = np.load(args.base_cic % i)
  proj = build_sky_proj(1+d, dmin=args.depth_min,dmax=args.depth_max,iid=args.iid)
  proj /= (args.depth_max-args.depth_min)

  hp.write_map("skymaps/proj_map_%d.fits" % i, proj)

  print proj.min(), proj.max()
  hp.mollview(proj, fig=1, coord='CG', cmap=plt.cm.copper, title='Sample %d' % i, min=args.minval, max=args.maxval)
  if args.proj_cat:
    hp.projscatter(b[idx], l[idx], lw=0, color=[0.1,0.8,0.8], s=2.0, alpha=0.7)

  ff.savefig(args.base_fig % i)



from cpython cimport bool
from cython cimport view
from cython.parallel import prange, parallel
from libc.math cimport sin, cos, abs, floor, round, sqrt
import numpy as np
cimport numpy as npx
cimport cython
from copy cimport *

ctypedef npx.float64_t DTYPE_t
DTYPE=np.float64
FORMAT_DTYPE="d"

__all__=["project_cic","line_of_sight_projection","spherical_projection","DTYPE","interp3d","interp2d"]

cdef extern from "project_tool.hpp" namespace "":

  DTYPE_t compute_projection(DTYPE_t *vertex_value, DTYPE_t *u, DTYPE_t *u0, DTYPE_t rho) nogil

cdef extern from "openmp.hpp" namespace "CosmoTool":
  int smp_get_max_threads() nogil
  int smp_get_thread_id() nogil


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef void interp3d_INTERNAL_periodic(DTYPE_t x, DTYPE_t y, 
                                        DTYPE_t z, 
                                        DTYPE_t[:,:,:] d, DTYPE_t Lbox, DTYPE_t *retval) nogil:

    cdef int Ngrid = d.shape[0]
    cdef DTYPE_t inv_delta = Ngrid/Lbox
    cdef int ix, iy, iz
    cdef DTYPE_t f[2][2][2]
    cdef DTYPE_t rx, ry, rz
    cdef int jx, jy, jz

    rx = (inv_delta*x)
    ry = (inv_delta*y)
    rz = (inv_delta*z)

    ix = int(floor(rx))
    iy = int(floor(ry))
    iz = int(floor(rz))


    rx -= ix
    ry -= iy
    rz -= iz

    ix = ix % Ngrid
    iy = iy % Ngrid
    iz = iz % Ngrid

    jx = (ix+1)%Ngrid
    jy = (iy+1)%Ngrid
    jz = (iz+1)%Ngrid

    ix = ix%Ngrid
    iy = iy%Ngrid
    iz = iz%Ngrid

    f[0][0][0] = (1-rx)*(1-ry)*(1-rz)
    f[1][0][0] = (  rx)*(1-ry)*(1-rz)
    f[0][1][0] = (1-rx)*(  ry)*(1-rz)
    f[1][1][0] = (  rx)*(  ry)*(1-rz)

    f[0][0][1] = (1-rx)*(1-ry)*(  rz)
    f[1][0][1] = (  rx)*(1-ry)*(  rz)
    f[0][1][1] = (1-rx)*(  ry)*(  rz)
    f[1][1][1] = (  rx)*(  ry)*(  rz)

    retval[0] = \
        d[ix ,iy ,iz ] * f[0][0][0] + \
        d[jx ,iy ,iz ] * f[1][0][0] + \
        d[ix ,jy ,iz ] * f[0][1][0] + \
        d[jx ,jy ,iz ] * f[1][1][0] + \
        d[ix ,iy ,jz ] * f[0][0][1] + \
        d[jx ,iy ,jz ] * f[1][0][1] + \
        d[ix ,jy ,jz ] * f[0][1][1] + \
        d[jx ,jy ,jz ] * f[1][1][1]

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef void ngp3d_INTERNAL_periodic(DTYPE_t x, DTYPE_t y, 
                                 DTYPE_t z, 
                                 DTYPE_t[:,:,:] d, DTYPE_t Lbox, DTYPE_t *retval) nogil:

    cdef int Ngrid = d.shape[0]
    cdef DTYPE_t inv_delta = Ngrid/Lbox
    cdef int ix, iy, iz
    cdef DTYPE_t f[2][2][2]
    cdef DTYPE_t rx, ry, rz
    cdef int jx, jy, jz

    rx = (inv_delta*x)
    ry = (inv_delta*y)
    rz = (inv_delta*z)

    ix = int(round(rx))
    iy = int(round(ry))
    iz = int(round(rz))


    ix = ix%Ngrid
    iy = iy%Ngrid
    iz = iz%Ngrid

    retval[0] = d[ix ,iy ,iz ] 


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef void ngp3d_INTERNAL(DTYPE_t x, DTYPE_t y, 
                         DTYPE_t z, 
                         DTYPE_t[:,:,:] d, DTYPE_t Lbox, DTYPE_t *retval, DTYPE_t inval) nogil:

    cdef int Ngrid = d.shape[0]
    cdef DTYPE_t inv_delta = Ngrid/Lbox
    cdef int ix, iy, iz
    cdef DTYPE_t f[2][2][2]
    cdef DTYPE_t rx, ry, rz
    cdef int jx, jy, jz

    rx = (inv_delta*x)
    ry = (inv_delta*y)
    rz = (inv_delta*z)

    ix = int(round(rx))
    iy = int(round(ry))
    iz = int(round(rz))

    if ((ix < 0)  or (ix+1) >= Ngrid or (iy < 0)  or (iy+1) >= Ngrid or (iz < 0)  or (iz+1) >= Ngrid):
      retval[0] = inval
      return

    retval[0] = d[ix ,iy ,iz ] 


@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cdef void interp3d_INTERNAL(DTYPE_t x, DTYPE_t y, 
                               DTYPE_t z, 
                               DTYPE_t[:,:,:] d, DTYPE_t Lbox, DTYPE_t *retval, DTYPE_t inval) nogil:
    
    cdef int Ngrid = d.shape[0]
    cdef DTYPE_t inv_delta = Ngrid/Lbox
    cdef int ix, iy, iz
    cdef DTYPE_t f[2][2][2]
    cdef DTYPE_t rx, ry, rz

    rx = (inv_delta*x)
    ry = (inv_delta*y)
    rz = (inv_delta*z)

    ix = int(floor(rx))
    iy = int(floor(ry))
    iz = int(floor(rz))

    rx -= ix
    ry -= iy
    rz -= iz

    if ((ix < 0)  or (ix+1) >= Ngrid or (iy < 0)  or (iy+1) >= Ngrid or (iz < 0)  or (iz+1) >= Ngrid):
      retval[0] = inval
      return
    #   assert ((ix >= 0) and ((ix+1) < Ngrid))
#    assert ((iy >= 0) and ((iy+1) < Ngrid))
#    assert ((iz >= 0) and ((iz+1) < Ngrid))

    f[0][0][0] = (1-rx)*(1-ry)*(1-rz)
    f[1][0][0] = (  rx)*(1-ry)*(1-rz)
    f[0][1][0] = (1-rx)*(  ry)*(1-rz)
    f[1][1][0] = (  rx)*(  ry)*(1-rz)

    f[0][0][1] = (1-rx)*(1-ry)*(  rz)
    f[1][0][1] = (  rx)*(1-ry)*(  rz)
    f[0][1][1] = (1-rx)*(  ry)*(  rz)
    f[1][1][1] = (  rx)*(  ry)*(  rz)

    retval[0] = \
        d[ix  ,iy  ,iz  ] * f[0][0][0] + \
        d[ix+1,iy  ,iz  ] * f[1][0][0] + \
        d[ix  ,iy+1,iz  ] * f[0][1][0] + \
        d[ix+1,iy+1,iz  ] * f[1][1][0] + \
        d[ix  ,iy  ,iz+1] * f[0][0][1] + \
        d[ix+1,iy  ,iz+1] * f[1][0][1] + \
        d[ix  ,iy+1,iz+1] * f[0][1][1] + \
        d[ix+1,iy+1,iz+1] * f[1][1][1]

@cython.boundscheck(False)
def interp3d(x not None, y not None, 
             z not None,
             npx.ndarray[DTYPE_t, ndim=3] d not None, DTYPE_t Lbox,
             bool periodic=False, bool centered=True, bool ngp=False, DTYPE_t inval = 0):
    """ interp3d(x,y,z,d,Lbox,periodic=False,centered=True,ngp=False) -> interpolated values
    
    Compute the tri-linear interpolation of the given field (d) at the given position (x,y,z). It assumes that they are box-centered coordinates. So (x,y,z) == (0,0,0) is equivalent to the pixel at (Nx/2,Ny/2,Nz/2) with Nx,Ny,Nz = d.shape. If periodic is set, it assumes the box is periodic 
"""
    cdef npx.ndarray[DTYPE_t] out
    cdef DTYPE_t[:] out_slice
    cdef DTYPE_t[:] ax, ay, az
    cdef DTYPE_t[:,:,:] in_slice
    cdef DTYPE_t retval
    cdef long i
    cdef long Nelt
    cdef int myperiodic, myngp
    cdef DTYPE_t shifter

    myperiodic = periodic
    myngp = ngp

    if centered:
      shifter = Lbox/2
    else:
      shifter = 0

    if d.shape[0] != d.shape[1] or d.shape[0] != d.shape[2]:
      raise ValueError("Grid must have a cubic shape")


    ierror = IndexError("Interpolating outside range")
    if type(x) == np.ndarray or type(y) == np.ndarray or type(z) == np.ndarray:
        if type(x) != np.ndarray or type(y) != np.ndarray or type(z) != np.ndarray:
            raise ValueError("All or no array. No partial arguments")
        
        ax = x
        ay = y
        az = z
        assert ax.size == ay.size and ax.size == az.size
        
        out = np.empty(x.shape, dtype=DTYPE)
        out_slice = out
        in_slice = d
        Nelt = ax.size
        with nogil:
          if not myngp:
            if myperiodic:
              for i in prange(Nelt):
                interp3d_INTERNAL_periodic(shifter+ax[i], shifter+ay[i], shifter+az[i], in_slice, Lbox, &out_slice[i])
            else:
              for i in prange(Nelt):
                interp3d_INTERNAL(shifter+ax[i], shifter+ay[i], shifter+az[i], in_slice, Lbox, &out_slice[i], inval)
          else:
            if myperiodic:
              for i in prange(Nelt):
                ngp3d_INTERNAL_periodic(shifter+ax[i], shifter+ay[i], shifter+az[i], in_slice, Lbox, &out_slice[i])
            else:
              for i in prange(Nelt):
                ngp3d_INTERNAL(shifter+ax[i], shifter+ay[i], shifter+az[i], in_slice, Lbox, &out_slice[i], inval)
        return out
    else:
      if not myngp:
        if periodic:
          interp3d_INTERNAL_periodic(shifter+x, shifter+y, shifter+z, d, Lbox, &retval)
        else:
          interp3d_INTERNAL(shifter+x, shifter+y, shifter+z, d, Lbox, &retval, inval)
      else:
        if periodic:
          ngp3d_INTERNAL_periodic(shifter+x, shifter+y, shifter+z, d, Lbox, &retval)
        else:
          ngp3d_INTERNAL(shifter+x, shifter+y, shifter+z, d, Lbox, &retval, inval)
      return retval

@cython.boundscheck(False)
@cython.cdivision(True)
cdef DTYPE_t interp2d_INTERNAL_periodic(DTYPE_t x, DTYPE_t y,
                                        npx.ndarray[DTYPE_t, ndim=2] d, DTYPE_t Lbox) except? 0:

    cdef int Ngrid = d.shape[0]
    cdef DTYPE_t inv_delta = Ngrid/Lbox
    cdef int ix, iy
    cdef DTYPE_t f[2][2]
    cdef DTYPE_t rx, ry
    cdef int jx, jy

    rx = (inv_delta*x + Ngrid/2)
    ry = (inv_delta*y + Ngrid/2)
    
    ix = int(floor(rx))
    iy = int(floor(ry))
    
    rx -= ix
    ry -= iy

    while ix < 0:
      ix += Ngrid
    while iy < 0:
      iy += Ngrid
    
    jx = (ix+1)%Ngrid
    jy = (iy+1)%Ngrid
    
    assert ((ix >= 0) and ((jx) < Ngrid))
    assert ((iy >= 0) and ((jy) < Ngrid))
    
    f[0][0] = (1-rx)*(1-ry)
    f[1][0] = (  rx)*(1-ry)
    f[0][1] = (1-rx)*(  ry)
    f[1][1] = (  rx)*(  ry)

    return \
        d[ix ,iy ] * f[0][0] + \
        d[jx ,iy ] * f[1][0] + \
        d[ix ,jy ] * f[0][1] + \
        d[jx ,jy ] * f[1][1]


@cython.boundscheck(False)
@cython.cdivision(True)
cdef DTYPE_t interp2d_INTERNAL(DTYPE_t x, DTYPE_t y,
                               npx.ndarray[DTYPE_t, ndim=2] d, DTYPE_t Lbox) except? 0:
    
    cdef int Ngrid = d.shape[0]
    cdef DTYPE_t inv_delta = Ngrid/Lbox
    cdef int ix, iy
    cdef DTYPE_t f[2][2]
    cdef DTYPE_t rx, ry

    rx = (inv_delta*x + Ngrid/2)
    ry = (inv_delta*y + Ngrid/2)

    ix = int(floor(rx))
    iy = int(floor(ry))

    rx -= ix
    ry -= iy

    if ((ix < 0)  or (ix+1) >= Ngrid):
      raise IndexError("X coord out of bound (ix=%d, x=%g)" % (ix,x))
    if ((iy < 0)  or (iy+1) >= Ngrid):
      raise IndexError("Y coord out of bound (iy=%d, y=%g)" % (iy,y))
#   assert ((ix >= 0) and ((ix+1) < Ngrid))
#    assert ((iy >= 0) and ((iy+1) < Ngrid))
#    assert ((iz >= 0) and ((iz+1) < Ngrid))

    f[0][0] = (1-rx)*(1-ry)
    f[1][0] = (  rx)*(1-ry)
    f[0][1] = (1-rx)*(  ry)
    f[1][1] = (  rx)*(  ry)

    return \
        d[ix  ,iy  ] * f[0][0] + \
        d[ix+1,iy  ] * f[1][0] + \
        d[ix  ,iy+1] * f[0][1] + \
        d[ix+1,iy+1] * f[1][1]
        
def interp2d(x not None, y not None,
             npx.ndarray[DTYPE_t, ndim=2] d not None, DTYPE_t Lbox,
             bool periodic=False):
    cdef npx.ndarray[DTYPE_t] out
    cdef npx.ndarray[DTYPE_t] ax, ay
    cdef int i

    if d.shape[0] != d.shape[1]:
      raise ValueError("Grid must have a square shape")

    if type(x) == np.ndarray or type(y) == np.ndarray:
        if type(x) != np.ndarray or type(y) != np.ndarray:
            raise ValueError("All or no array. No partial arguments")
        
        ax = x
        ay = y
        assert ax.size == ay.size 
        
        out = np.empty(x.shape, dtype=DTYPE)
        if periodic:
          for i in range(ax.size):
            out[i] = interp2d_INTERNAL_periodic(ax[i], ay[i], d, Lbox)
        else:
          for i in range(ax.size):
            out[i] = interp2d_INTERNAL(ax[i], ay[i], d, Lbox)

        return out
    else:
      if periodic:
        return interp2d_INTERNAL_periodic(x, y, d, Lbox)
      else:
        return interp2d_INTERNAL(x, y, d, Lbox)
        
 
@cython.boundscheck(False)
@cython.cdivision(True)
cdef void INTERNAL_project_cic_no_mass(DTYPE_t[:,:,:] g,
                                       DTYPE_t[:,:] x, int Ngrid, double Lbox, double shifter) nogil:
    cdef double delta_Box = Ngrid/Lbox
    cdef int i
    cdef double a[3]
    cdef double c[3]
    cdef int b[3]
    cdef int do_not_put

    for i in range(x.shape[0]):

        do_not_put = 0
        for j in range(3):
            a[j] = (x[i,j]+shifter)*delta_Box
            b[j] = int(floor(a[j]))
            a[j] -= b[j]
            c[j] = 1-a[j]
            if b[j] < 0 or b[j]+1 >= Ngrid:
              do_not_put = True

        if not do_not_put:
            g[b[0],b[1],b[2]] += c[0]*c[1]*c[2]
            g[b[0]+1,b[1],b[2]] += a[0]*c[1]*c[2]
            g[b[0],b[1]+1,b[2]] += c[0]*a[1]*c[2]
            g[b[0]+1,b[1]+1,b[2]] += a[0]*a[1]*c[2]

            g[b[0],b[1],b[2]+1] += c[0]*c[1]*a[2]
            g[b[0]+1,b[1],b[2]+1] += a[0]*c[1]*a[2]
            g[b[0],b[1]+1,b[2]+1] += c[0]*a[1]*a[2]
            g[b[0]+1,b[1]+1,b[2]+1] += a[0]*a[1]*a[2]

@cython.boundscheck(False)
@cython.cdivision(True)
cdef void INTERNAL_project_cic_no_mass_periodic(DTYPE_t[:,:,:] g,
                                                DTYPE_t[:,:] x, int Ngrid, double Lbox, double shifter) nogil:
    cdef double delta_Box = Ngrid/Lbox
    cdef int i
    cdef double a[3]
    cdef double c[3]
    cdef int b[3]
    cdef int b1[3]
    cdef int do_not_put
    cdef DTYPE_t[:,:] ax
    cdef DTYPE_t[:,:,:] ag

    ax = x
    ag = g

    for i in range(x.shape[0]):

        do_not_put = 0
        for j in range(3):
            a[j] = (ax[i,j]+shifter)*delta_Box
            b[j] = int(floor(a[j]))
            b1[j] = (b[j]+1) % Ngrid

            a[j] -= b[j]
            c[j] = 1-a[j]

            b[j] %= Ngrid

        ag[b[0],b[1],b[2]] += c[0]*c[1]*c[2]
        ag[b1[0],b[1],b[2]] += a[0]*c[1]*c[2]
        ag[b[0],b1[1],b[2]] += c[0]*a[1]*c[2]
        ag[b1[0],b1[1],b[2]] += a[0]*a[1]*c[2]
        
        ag[b[0],b[1],b1[2]] += c[0]*c[1]*a[2]
        ag[b1[0],b[1],b1[2]] += a[0]*c[1]*a[2]
        ag[b[0],b1[1],b1[2]] += c[0]*a[1]*a[2]
        ag[b1[0],b1[1],b1[2]] += a[0]*a[1]*a[2]


@cython.boundscheck(False)
@cython.cdivision(True)
cdef void INTERNAL_project_cic_with_mass(DTYPE_t[:,:,:] g,
                                         DTYPE_t[:,:] x,
                                         DTYPE_t[:] mass,
                                         int Ngrid, double Lbox, double shifter) nogil:
    cdef double delta_Box = Ngrid/Lbox
    cdef int i
    cdef double a[3]
    cdef double c[3]
    cdef DTYPE_t m0
    cdef int b[3]

    for i in range(x.shape[0]):

        do_not_put = False
        for j in range(3):
            a[j] = (x[i,j]+shifter)*delta_Box
            b[j] = int(a[j])
            a[j] -= b[j]
            c[j] = 1-a[j]
            if b[j] < 0 or b[j]+1 >= Ngrid:
              do_not_put = True

        if not do_not_put:
            m0 = mass[i]

            g[b[0],b[1],b[2]] += c[0]*c[1]*c[2]*m0
            g[b[0]+1,b[1],b[2]] += a[0]*c[1]*c[2]*m0
            g[b[0],b[1]+1,b[2]] += c[0]*a[1]*c[2]*m0
            g[b[0]+1,b[1]+1,b[2]] += a[0]*a[1]*c[2]*m0

            g[b[0],b[1],b[2]+1] += c[0]*c[1]*a[2]*m0
            g[b[0]+1,b[1],b[2]+1] += a[0]*c[1]*a[2]*m0
            g[b[0],b[1]+1,b[2]+1] += c[0]*a[1]*a[2]*m0
            g[b[0]+1,b[1]+1,b[2]+1] += a[0]*a[1]*a[2]*m0

@cython.boundscheck(False)
@cython.cdivision(True)
cdef void INTERNAL_project_cic_with_mass_periodic(DTYPE_t[:,:,:] g,
                                                  DTYPE_t[:,:] x,
                                                  DTYPE_t[:] mass,
                                                  int Ngrid, double Lbox, double shifter) nogil:
    cdef double half_Box = 0.5*Lbox, m0
    cdef double delta_Box = Ngrid/Lbox
    cdef int i
    cdef double a[3]
    cdef double c[3]
    cdef int b[3]
    cdef int b1[3]

    for i in range(x.shape[0]):
        for j in range(3):
            a[j] = (x[i,j]+shifter)*delta_Box
            b[j] = int(floor(a[j]))
            b1[j] = (b[j]+1) % Ngrid

            a[j] -= b[j]
            c[j] = 1-a[j]

            b[j] %= Ngrid


        m0 = mass[i]
        g[b[0],b[1],b[2]] += c[0]*c[1]*c[2]*m0
        g[b1[0],b[1],b[2]] += a[0]*c[1]*c[2]*m0
        g[b[0],b1[1],b[2]] += c[0]*a[1]*c[2]*m0
        g[b1[0],b1[1],b[2]] += a[0]*a[1]*c[2]*m0
        
        g[b[0],b[1],b1[2]] += c[0]*c[1]*a[2]*m0
        g[b1[0],b[1],b1[2]] += a[0]*c[1]*a[2]*m0
        g[b[0],b1[1],b1[2]] += c[0]*a[1]*a[2]*m0
        g[b1[0],b1[1],b1[2]] += a[0]*a[1]*a[2]*m0

       
def project_cic(npx.ndarray[DTYPE_t, ndim=2] x not None, npx.ndarray[DTYPE_t, ndim=1] mass, int Ngrid,
                double Lbox, bool periodic = False, centered=True):
    """
    project_cic(x array (N,3), mass (may be None), Ngrid, Lbox, periodict, centered=True)

    This function does a Cloud-In-Cell projection of a 3d unstructured dataset. First argument is a Nx3 array of coordinates. 
    Second argument is an optinal mass. Ngrid is the size output grid and Lbox is the physical size of the grid.
    """
    cdef npx.ndarray[DTYPE_t, ndim=3] g
    cdef double shifter
    cdef bool local_periodic
    cdef DTYPE_t[:] d_mass

    local_periodic = periodic

    if centered:
      shifter = 0.5*Lbox
    else:
      shifter = 0

    if x.shape[1] != 3:
        raise ValueError("Invalid shape for x array")

    if mass is not None and mass.shape[0] != x.shape[0]:
        raise ValueError("Mass array and coordinate array must have the same number of elements")

    g = np.zeros((Ngrid,Ngrid,Ngrid),dtype=DTYPE)
    cdef DTYPE_t[:,:,:] d_g = g
    cdef DTYPE_t[:,:] d_x = x

    if not local_periodic:
       if mass is None:
         with nogil:
           INTERNAL_project_cic_no_mass(d_g, d_x, Ngrid, Lbox, shifter)
       else:
         d_mass = mass
         with nogil:
           INTERNAL_project_cic_with_mass(d_g, d_x, d_mass, Ngrid, Lbox, shifter)       
    else:
       if mass is None:
         with nogil:
          INTERNAL_project_cic_no_mass_periodic(d_g, d_x, Ngrid, Lbox, shifter)
       else:
         d_mass = mass
         with nogil:
          INTERNAL_project_cic_with_mass_periodic(d_g, d_x, d_mass, Ngrid, Lbox, shifter)       
         
    return g

def tophat_fourier_internal(npx.ndarray[DTYPE_t, ndim=1] x not None):
    cdef int i
    cdef npx.ndarray[DTYPE_t] y    
    cdef DTYPE_t x0

    y = np.empty(x.size, dtype=DTYPE)

    for i in range(x.size):
        x0 = x[i]
        if abs(x0)<1e-5:
            y[i] = 1
        else:
            y[i] = (3*(sin(x0) - x0 * cos(x0))/(x0**3))

    return y

def tophat_fourier(x not None):
    cdef npx.ndarray[DTYPE_t, ndim=1] b

    if type(x) != np.ndarray:
        raise ValueError("x must be a Numpy array")

    b = np.array(x, dtype=DTYPE).ravel()

    b = tophat_fourier_internal(b)

    return b.reshape(x.shape)

    

@cython.boundscheck(False)
@cython.cdivision(True)
cdef DTYPE_t cube_integral(DTYPE_t u[3], DTYPE_t u0[3], int r[1], DTYPE_t alpha_max) nogil:
    cdef DTYPE_t tmp_a
    cdef DTYPE_t v[3]
    cdef int i, j

    for i in xrange(3):
        if u[i] == 0.:
            continue

        if u[i] < 0:
            tmp_a = -u0[i]/u[i]
        else:
            tmp_a = (1-u0[i])/u[i]

        if tmp_a < alpha_max:
            alpha_max = tmp_a
            j = i

    for i in range(3):
        u0[i] += u[i]*alpha_max

    r[0] = j

    return alpha_max

@cython.boundscheck(False)
@cython.cdivision(True)
cdef DTYPE_t cube_integral_trilin(DTYPE_t u[3], DTYPE_t u0[3], int r[1], DTYPE_t vertex_value[8], DTYPE_t alpha_max) nogil:
    cdef DTYPE_t I, tmp_a
    cdef DTYPE_t v[3]
    cdef DTYPE_t term[4]
    cdef int i, j, q

    j = 0
    for i in range(3):
        if u[i] == 0.:
            continue

        if u[i] < 0:
            tmp_a = -u0[i]/u[i]
        else:
            tmp_a = (1-u0[i])/u[i]

        if tmp_a < alpha_max:
            alpha_max = tmp_a
            j = i
     
    I = compute_projection(vertex_value, u, u0, alpha_max)
    
    for i in xrange(3):
        u0[i] += u[i]*alpha_max

    # alpha_max is the integration length
    # we integrate between 0 and alpha_max (curvilinear coordinates)
    r[0] = j
    
    return I

@cython.boundscheck(False)
cdef DTYPE_t integrator0(DTYPE_t[:,:,:] density,
                         DTYPE_t u[3], DTYPE_t u0[3], int u_delta[3], int iu0[3], int jumper[1], DTYPE_t alpha_max) nogil:
    cdef DTYPE_t d
    
    d = density[iu0[0], iu0[1], iu0[2]]
    
    return cube_integral(u, u0, jumper, alpha_max)*d

@cython.boundscheck(False)
cdef DTYPE_t integrator1(DTYPE_t[:,:,:] density,
                         DTYPE_t u[3], DTYPE_t u0[3], int u_delta[3], int iu0[3], int jumper[1], DTYPE_t alpha_max) nogil:
    cdef DTYPE_t vertex_value[8]
    cdef DTYPE_t d
    cdef int a[3][2]
    cdef int i
    
    for i in xrange(3):
      a[i][0] = iu0[i]
      a[i][1] = iu0[i]+1

    vertex_value[0 + 2*0 + 4*0] = density[a[0][0], a[1][0], a[2][0]]
    vertex_value[1 + 2*0 + 4*0] = density[a[0][1], a[1][0], a[2][0]]
    vertex_value[0 + 2*1 + 4*0] = density[a[0][0], a[1][1], a[2][0]]
    vertex_value[1 + 2*1 + 4*0] = density[a[0][1], a[1][1], a[2][0]]

    vertex_value[0 + 2*0 + 4*1] = density[a[0][0], a[1][0], a[2][1]]
    vertex_value[1 + 2*0 + 4*1] = density[a[0][1], a[1][0], a[2][1]]
    vertex_value[0 + 2*1 + 4*1] = density[a[0][0], a[1][1], a[2][1]]
    vertex_value[1 + 2*1 + 4*1] = density[a[0][1], a[1][1], a[2][1]]

    return cube_integral_trilin(u, u0, jumper, vertex_value, alpha_max)


    
@cython.boundscheck(False)
cdef DTYPE_t C_line_of_sight_projection(DTYPE_t[:,:,:] density,
                             DTYPE_t a_u[3],
                             DTYPE_t min_distance,
                             DTYPE_t max_distance, DTYPE_t[:] shifter, int integrator_id) nogil except? 0:

    cdef DTYPE_t u[3] 
    cdef DTYPE_t ifu0[3]
    cdef DTYPE_t u0[3]
    cdef DTYPE_t utot[3]
    cdef int u_delta[3]
    cdef int iu0[3]
    cdef int i
    cdef int N = density.shape[0]
    cdef int half_N = density.shape[0]/2
    cdef int completed
    cdef DTYPE_t I0, d, dist2, delta, s, max_distance2
    cdef int jumper[1]
    
    cdef DTYPE_t (*integrator)(DTYPE_t[:,:,:],
                               DTYPE_t u[3], DTYPE_t u0[3], int u_delta[3], int iu0[3], int jumper[1], DTYPE_t alpha_max) nogil

    if integrator_id == 0:
      integrator = integrator0
    else:
      integrator = integrator1

    max_distance2 = max_distance**2

    for i in range(3):
        u[i] = a_u[i]
        u0[i] = a_u[i]*min_distance
        ifu0[i] = half_N+u0[i]+shifter[i]
        if (ifu0[i] <= 0 or ifu0[i] >= N):
          return 0
        iu0[i] = int(floor(ifu0[i]))
        u0[i] = ifu0[i]-iu0[i]
        u_delta[i] = 1 if iu0[i] > 0 else -1
        if (not ((iu0[i]>= 0) and (iu0[i] < N))):
          with gil:
            raise RuntimeError("iu0[%d] = %d !!" % (i,iu0[i]))
            
        if (not (u0[i]>=0 and u0[i]<=1)):
          with gil:
            raise RuntimeError("u0[%d] = %g !" % (i,u0[i]))

    completed = 0
    if ((iu0[0] >= N-1) or (iu0[0] <= 0) or
        (iu0[1] >= N-1) or (iu0[1] <= 0) or
        (iu0[2] >= N-1) or (iu0[2] <= 0)):
      completed = 1 

    I0 = 0
    jumper[0] = 0
    dist2 = 0
    while completed == 0:
        I0 += integrator(density, u, u0, u_delta, iu0, jumper, max_distance-sqrt(dist2))

        if u[jumper[0]] < 0:
            iu0[jumper[0]] -= 1
            u0[jumper[0]] = 1
        else:
            iu0[jumper[0]] += 1
            u0[jumper[0]] = 0

            
        if ((iu0[0] >= N-1) or (iu0[0] <= 0) or 
            (iu0[1] >= N-1) or (iu0[1] <= 0) or
            (iu0[2] >= N-1) or (iu0[2] <= 0)):
            completed = 1
        else:
          dist2 = 0
          for i in range(3):
             delta =  iu0[i]+u0[i]-half_N-shifter[i]
             dist2 += delta*delta

          if (dist2 > max_distance2):
            # Remove the last portion of the integral
            #delta = sqrt(dist2) - max_distance
            #I0 -= d*delta
            completed = 1
            
    return I0

def line_of_sight_projection(DTYPE_t[:,:,:] density not None,
                             DTYPE_t[:] a_u not None,
                             DTYPE_t min_distance,
                             DTYPE_t max_distance, DTYPE_t[:] shifter not None, int integrator_id=0):
  cdef DTYPE_t u[3]
  
  u[0] = a_u[0]
  u[1] = a_u[1]
  u[2] = a_u[2]
  
  return C_line_of_sight_projection(density,
                             u,
                             min_distance,
                             max_distance, shifter, integrator_id)

cdef double _spherical_projloop(double theta, double phi, DTYPE_t[:,:,:] density, 
                double min_distance, double max_distance, 
                DTYPE_t[:] shifter, int integrator_id) nogil:
  cdef DTYPE_t u0[3]

  stheta = sin(theta)
  u0[0] = cos(phi)*stheta
  u0[1] = sin(phi)*stheta
  u0[2] = cos(theta)
 
  return C_line_of_sight_projection(density, u0, min_distance, max_distance, shifter, integrator_id)


@cython.boundscheck(False)
def spherical_projection(int Nside,
                         npx.ndarray[DTYPE_t, ndim=3] density not None,
                         DTYPE_t min_distance,
                         DTYPE_t max_distance, int progress=1, int integrator_id=0, DTYPE_t[:] shifter = None, int booster=-1):
    """
    spherical_projection(Nside, density, min_distance, max_distance, progress=1, integrator_id=0, shifter=None, booster=-1)
    
    Keyword arguments:
      progress (int): show progress if it is equal to 1
      integrator_id (int): specify the order of integration along the line of shift
      shifter (DTYPE_t array): this is an array of size 3. It specifies the amount of shift to apply to the center, in unit of voxel
      booster (int): what is the frequency of refreshment of the progress bar. Small number decreases performance by locking the GIL.
      

    Arguments:
      Nside (int): Nside of the returned map
      density (NxNxN array): this is the density field, expressed as a cubic array
      min_distance (float): lower bound of the integration
      max_distance (float): upper bound of the integration
      
    Returns:
      an healpix map, as a 1-dimensional array.
    """
    import healpy as hp
    import progressbar as pb
    cdef int i
    cdef DTYPE_t[:] theta,phi
    cdef DTYPE_t[:,:,:] density_view
    cdef DTYPE_t[:] outm
    cdef int[:] job_done
    cdef npx.ndarray[DTYPE_t, ndim=1] outm_array
    cdef long N, N0
    cdef double stheta
    cdef int tid
    
    if shifter is None:
        shifter = view.array(shape=(3,), format=FORMAT_DTYPE, itemsize=sizeof(DTYPE_t))
        shifter[:] = 0
        
    print("allocating map")
    outm_array = np.empty(hp.nside2npix(Nside),dtype=DTYPE)
    print("initializing views")
    outm = outm_array
    density_view = density

    print("progress?")
    if progress != 0:
      p = pb.ProgressBar(maxval=outm.size,widgets=[pb.Bar(), pb.ETA()]).start()

    N = smp_get_max_threads()
    N0 = outm.size
    
    if booster < 0:
      booster = 1#000
    
    job_done = view.array(shape=(N,), format="i", itemsize=sizeof(int))
    job_done[:] = 0
    theta,phi = hp.pix2ang(Nside, np.arange(N0))
    with nogil, parallel():
      tid = smp_get_thread_id()
      for i in prange(N0,schedule='dynamic',chunksize=256):
          if progress != 0 and (i%booster) == 0:
            with gil:
              p.update(_mysum(job_done))
          outm[i] = _spherical_projloop(theta[i], phi[i], density_view, min_distance, max_distance, shifter, integrator_id)
          job_done[tid] += 1

    if progress:
      p.finish()


    return outm_array



import warnings
from _cosmotool import *

class SimulationBare(PySimulationBase):

  def __init__(self, *args):
    if len(args) == 0:
      return
      
    if not isinstance(args[0], PySimulationBase):
      raise TypeError("Simulation object to mirror must be a PySimulationBase")
    s = args[0]

    self.positions = [q.copy() for q in s.getPositions()] if s.getPositions() is not None else None
    self.velocities = [q.copy() for q in s.getVelocities()] if s.getVelocities() is not None else None
    self.identifiers = s.getIdentifiers().copy() if s.getIdentifiers() is not None else None
    self.types = s.getTypes().copy() if s.getTypes() is not None else None
    self.boxsize = s.getBoxsize()
    self.time = s.getTime()
    self.Hubble = s.getHubble()
    self.Omega_M = s.getOmega_M()
    self.Omega_Lambda = s.getOmega_Lambda()
    try:
      self.masses = s.getMasses().copy() if s.getMasses() is not None else None
    except Exception as e:
      warnings.warn("Unexpected exception: " + repr(e))
  
  
  def merge(self, other):

    def _safe_merge(a, b):
      if b is not None:
        if a is not None:
          a = [np.append(q, r) for q,r in zip(a,b)]
        else:
          a = b
      return a
      
    def _safe_merge0(a, b):
      if b is not None:
        if a is not None:
          a = np.append(a, b)
        else:
          a = b
      return a
  
  
    assert self.time == other.getTime()
    assert self.Hubble == other.getHubble()
    assert self.boxsize == other.getBoxsize()
    assert self.Omega_M == other.getOmega_M()
    assert self.Omega_Lambda == other.getOmega_Lambda()
    
    self.positions = _safe_merge(self.positions, other.getPositions())
    self.velocities = _safe_merge(self.velocities, other.getVelocities())
    self.identifiers = _safe_merge0(self.identifiers, other.getIdentifiers())
    self.types = _safe_merge0(self.types, other.getTypes())
    try:
      self.masses = _safe_merge0(self.masses, other.getMasses())
    except Exception as e:
      warnings.warn("Unexpected exception: " + repr(e));
      self.masses = None

  def getTypes(self):
    return self.types
    
  def getPositions(self):
    return self.positions
    
  def getVelocities(self):
    return self.velocities
    
  def getIdentifiers(self):
    return self.identifiers

  def getMasses(self):
    return self.masses

  def getTime(self):
    return self.time
    
  def getHubble(self):
    return self.Hubble

  def getBoxsize(self):
    return self.boxsize

  def getOmega_M(self):
    return self.Omega_M

  def getOmega_Lambda(self):
    return self.Omega_Lambda


def simpleWriteGadget(filename, positions, boxsize=1.0, Hubble=100, Omega_M=0.30, time=1, velocities=None, identifiers=None):

  s = SimulationBare()
  
  s.positions = positions
  
  if velocities is not None:
    s.velocities = velocities
  else:
    s.velocities = [np.zeros(positions[0].size,dtype=np.float32)]*3
    
  if identifiers is not None:
    s.identifiers = identifiers
  else:
    s.identifiers = np.arange(positions[0].size, dtype=np.int64)
    
  s.Hubble = Hubble
  s.time = time
  s.Omega_M = Omega_M
  s.Omega_Lambda = 1-Omega_M
  s.boxsize = boxsize
    
  writeGadget(filename, s)
  
def loadRamsesAll(basepath, snapshot_id, **kwargs):
  """This function loads an entire ramses snapshot in memory. The keyword arguments are the one accepted
  by cosmotool.loadRamses
  
  Args:
    basepath (str): The base path of the snapshot (i.e. the directory holding the output_XXXXX directories)
    snapshot_id (int): The identifier of the snapshot to load.

  Keyword args:
    verboseMulti (bool): If true, print some progress information on loading multiple files
    
  See Also:
    cosmotool.loadRamses
    
  """
  cpu_id = 0
  output = None
  verbose = kwargs.get('verboseMulti',False)
  new_kw = dict(kwargs)
  if 'verboseMulti' in new_kw:
    del new_kw['verboseMulti']
  while True:
    base = "%s/output_%05d" % (basepath,snapshot_id)
    if verbose:
      print("Loading sub-snapshot %s (cpu_id=%d)" % (base,cpu_id))
    s = loadRamses(base, snapshot_id, cpu_id, **new_kw)
    if s == None:
      break
    if output == None:
      output = SimulationBare(s)
    else:
      output.merge(s)
      
    cpu_id += 1

  return output 



import numpy as np
import cosmotool as ct
import borgicgen as bic
from matplotlib import pyplot as plt

cosmo={'omega_M_0':0.3175, 'h':0.6711}
cosmo['omega_lambda_0']=1-cosmo['omega_M_0']
cosmo['omega_k_0'] = 0
cosmo['omega_B_0']=0.049
cosmo['SIGMA8']=0.8344
cosmo['ns']=0.9624

zstart=50
astart=1/(1.+zstart)
halfPixelShift=False

posx,vel,density,N,L,a_ic,cosmo = bic.run_generation("initial_condition_borg.dat", 0.001, astart, cosmo, supersample=1, shiftPixel=halfPixelShift, do_lpt2=False)

w1 = bic.whitify(density, L, cosmo, supergenerate=1)
w2 = bic.whitify(density, L, cosmo, supergenerate=2)

N = w1.shape[0]
Ns = w2.shape[0]

w1_hat = np.fft.rfftn(w1)*(L/N)**3
w2_hat = np.fft.rfftn(w2)*(L/Ns)**3

P1, b1, dev1 = bic.bin_power(np.abs(w1_hat)**2, L, range=(0,3),bins=150,dev=True)
P2, b2, dev2 = bic.bin_power(np.abs(w2_hat)**2, L, range=(0,3),bins=150,dev=True)

fig = plt.figure(1)
fig.clf()
plt.fill_between(b1, P1*(1-dev1), P1*(1+dev1), label='Supergen=1', color='b')
plt.fill_between(b2, P2*(1-dev2), P2*(1+dev2), label='Supergen=2', color='g', alpha=0.5)
ax = plt.gca()
ax.set_xscale('log')
plt.ylim(0.5,1.5)
plt.xlim(1e-2,4)
plt.axhline(1.0, color='red', lw=4.0)
plt.legend()
plt.show()



from ._cosmotool import *
from ._project import *
from ._cosmo_power import *
from ._cosmo_cic import *
from ._fast_interp import *
from ._cosmomath import *
from .grafic import writeGrafic, writeWhitePhase, readGrafic, readWhitePhase
from .borg import read_borg_vol
from .cic import cicParticles
try:
  import pyopencl
  from .cl_cic import cl_CIC_Density
except:
  print("No opencl support")

from .simu import loadRamsesAll, simpleWriteGadget, SimulationBare
from .timing import time_block, timeit, timeit_quiet
from .bispectrum import bispectrum, powerspectrum
from .smooth import smooth_particle_density

try:
  from .fftw import CubeFT
except ImportError:
  print("No FFTW support")



import time                                                
from contextlib import contextmanager

@contextmanager
def time_block(name):
    """
    This generator measure the time taken by a step, and prints the result
    in second to the console.
    
    Arguments:
        name (str): prefix to print
    """
    ts = time.time()
    yield
    te = time.time()
 
    print('%s %2.2f sec' %  (name, te-ts))

def timeit(method):
    """This decorator add a timing request for each call to the decorated function.
    
    Arguments:
        method (function): the method to decorate
    """

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        print('%r (%r, %r) %2.2f sec' % (method.__name__, args, kw, te-ts))
        return result

    return timed

def timeit_quiet(method):
    """This decorator add a timing request for each call to the decorated function.
    Same as cosmotool.timeit_ but is quieter by not printing the values of the arguments.
    
    Arguments:
        method (function): the method to decorate
    """

    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()

        print('%r %2.2f sec' % (method.__name__, te-ts))
        return result

    return timed




import pyfftw
import numpy as np
import cosmotool as ct
import borgicgen as bic
import pickle

with file("wisdom") as f:
  pyfftw.import_wisdom(pickle.load(f))

cosmo={'omega_M_0':0.3175, 'h':0.6711}
cosmo['omega_lambda_0']=1-cosmo['omega_M_0']
cosmo['omega_k_0'] = 0
cosmo['omega_B_0']=0.049
cosmo['SIGMA8']=0.8344
cosmo['ns']=0.9624

supergen=1
zstart=99
astart=1/(1.+zstart)
halfPixelShift=False
zero_fill=False

if __name__=="__main__":
    bic.write_icfiles(*bic.run_generation("initial_density_1872.dat", 0.001, astart, cosmo, supersample=1, shiftPixel=halfPixelShift, do_lpt2=False, supergenerate=supergen), supergenerate=1, zero_fill=zero_fill)



### 
### BORG code is from J. Jasche
###
import io
import numpy as np
from numpy import *
import os.path
import array
import glob

class BorgVolume(object):

    def __init__(self, density, ranges):
        self.density = density
        self.ranges = ranges

def build_filelist(fdir):
    #builds list of all borg density fields which may be distributed over several directories
    
    fname_0=glob.glob(fdir[0]+'initial_density_*')
    fname_1=glob.glob(fdir[0]+'final_density_*')
       
    fdir=fdir[1:] #eliminate first element
       
    for fd in fdir:
      fname_0=fname_0+glob.glob(fd+'initial_density_*')
      fname_1=fname_1+glob.glob(fd+'final_density_*')
	   
    return fname_0, fname_1

def read_borg_vol(BORGFILE):
    """ Reading routine for BORG data
    """

    openfile=open(BORGFILE,'rb')

    period=0
    N0=0
    N1=0
    N2=0

    xmin=0
    xmax=0

    ymin=0
    ymax=0

    zmin=0
    zmax=0
    nlines=0

    while True:
      line=openfile.readline()
      s=line.rstrip('\n')
      r=s.rsplit(' ')
    
      if size(r)==5 : 
        if r[0] =="define":
          if r[1]=="Lattice" :
            N0=int(r[2])
            N1=int(r[3])
            N2=int(r[4])
            
          if size(r)==11 :
            if r[4] =="BoundingBox":
              xmin=float(r[5])
              xmax=float(r[6])
              ymin=float(r[7])
              ymax=float(r[8])
              zmin=float(r[9])
              zmax=float(r[10].rstrip(','))
              
          if r[0]=='@1': break
  
    ranges=[]
    ranges.append(xmin)
    ranges.append(xmax)
    ranges.append(ymin)
    ranges.append(ymax)
    ranges.append(zmin)
    ranges.append(zmax)
    
    #now read data
    data=np.fromfile(openfile, '>f4')
    data=data.reshape(N2,N0,N1)
    return BorgVolume(data,ranges)

def read_spec( fname ):
    """ Reading routine for ARES spectrum samples
    """
    x,y=np.loadtxt( fname ,usecols=(0,1),unpack=True)

    return x , y
    
        
def read_bias_nmean( fname ):
    """ Reading routine for ARES bias data
    """
    x,b0,b1,nmean=np.loadtxt( fname ,usecols=(0,1,2,3),unpack=True)

    return x , b0, b1, nmean

def read_nmean( fname ):
    """ Reading routine for BORG bias data
    """
    x,nmean=np.loadtxt( fname ,usecols=(0,1),unpack=True)

    return x, nmean    
    
def get_grid_values(xx,data, ranges):
    """ return values at grid positions
    """
    xmin=ranges[0]
    xmax=ranges[1]
    
    ymin=ranges[2]
    ymax=ranges[3]
    
    zmin=ranges[4]
    zmax=ranges[5]
    
    Lx= xmax-xmin
    Ly= ymax-ymin
    Lz= zmax-zmin
        
    Nx=shape(data)[0]
    Ny=shape(data)[1]
    Nz=shape(data)[2]
    
    dx=Lx/float(Nx)
    dy=Ly/float(Ny)
    dz=Lz/float(Nz)
    
    idx=(xx[:,0]-xmin)/dx
    idy=(xx[:,1]-ymin)/dz
    idz=(xx[:,2]-zmin)/dy
    
    idx=idx.astype(int)
    idy=idy.astype(int)
    idz=idz.astype(int)
    
    idflag=np.where( (idx>-1)*(idx<Nx)*(idy>-1)*(idy<Ny)*(idz>-1)*(idz<Nz) )
        
    flag=[False]*len(xx)
    vals=[-999.]*len(xx)
    
    flag=np.array(flag)
    vals=np.array(vals)
    flag[idflag]=True
    vals[idflag]=data[idx[idflag],idy[idflag],idz[idflag]]
        
    return vals,flag    
    
def get_mean_density(fdir, smin, step):
    """ estimate ensemble mean
    """
    import progressbar as pb

    print('-'*60)
    print('Get 3D ensemble mean density field')
    print('-'*60)

    fname0 = fdir + 'initial_density_'+str(0)+'.dat'
    fname1 = fdir + 'final_density_'+str(0)+'.dat'
        
    MEAN0,ranges=read_borg_vol(fname0)
    MEAN0=MEAN0*0.;
    VAR0=copy(MEAN0)
    MEAN1=copy(MEAN0)
    VAR1=copy(MEAN0)
    norm=0.
    
    idat=smin
    
    fname0 = fdir + 'initial_density_'+str(idat)+'.dat'
    fname1 = fdir + 'final_density_'+str(idat)+'.dat'
        #and (idat<smin+1000)
    while((os.path.exists(fname0))):
      auxdata0,auxranges0=read_borg_vol(fname0)
      auxdata1,auxranges1=read_borg_vol(fname1)
      auxx0=auxdata0
      auxx1=auxdata1
      MEAN0+=auxx0
      VAR0+=auxx0**2
      MEAN1+=auxx1
      VAR1+=auxx1**2

      norm+=1
      idat+=step
      fname0 = fdir + 'initial_density_'+str(idat)+'.dat'
      fname1 = fdir + 'final_density_'+str(idat)+'.dat'
      del auxranges0
      del auxdata0
      del auxranges1
      del auxdata1
	    
    MEAN0/=norm
    VAR0/=norm
    VAR0-=MEAN0**2
    VAR0=sqrt(fabs(VAR0))
    
    MEAN1/=norm
    VAR1/=norm
    VAR1-=MEAN1**2
    VAR1=sqrt(fabs(VAR1))
    
    
    return MEAN0,VAR0,MEAN1,VAR1,ranges

def get_mean_density_fdir(fdir,init,steps):
    """ estimate ensemble mean
    """
    import progressbar as pb

    print('-'*60)
    print('Get 3D ensemble mean density field')
    print('-'*60)

    fname0,fname1=build_filelist(fdir)
    
    fname0=fname0[init::steps]
    fname1=fname1[init::steps]
    
    borg=read_borg_vol(fname0[0])
    MEAN0 = borg.density
    RANGES0 = borg.ranges
    MEAN0=MEAN0*0.;
    VAR0=copy(MEAN0)
    MEAN1=copy(MEAN0)
    VAR1=copy(MEAN0)
    norm0=0.
    norm1=0.
    
    for fn in pb.ProgressBar(len(fname0))(fname0):
      auxborg=read_borg_vol(fn)
      auxdata0 = auxborg.density
      MEAN0+=auxdata0
      VAR0+=auxdata0**2.
      norm0+=1.
      del auxdata0
      del auxborg

    for fn in pb.ProgressBar(len(fname1))(fname1):
      auxborg1=read_borg_vol(fn)
      auxdata1 = auxborg1.density
      MEAN1+=auxdata1
      VAR1+=auxdata1**2.
      norm1+=1.
      del auxdata1
      del auxborg1
	
    MEAN0/=norm0
    VAR0/=norm0
    VAR0-=MEAN0**2
    VAR0=sqrt(fabs(VAR0))
    
    MEAN1/=norm1
    VAR1/=norm1
    VAR1-=MEAN1**2
    VAR1=sqrt(fabs(VAR1))
    
    
    return MEAN0,VAR0,MEAN1,VAR1,ranges    
    
    



from .timing import time_block as time_block_orig
import numpy as np
import pyopencl as cl
import pyopencl.array as cl_array
from contextlib import contextmanager

TIMER_ACTIVE=False

@contextmanager
def time_block_dummy(*args):
  yield

if TIMER_ACTIVE:
  time_block=time_block_orig
else: 
  time_block=time_block_dummy

CIC_PREKERNEL='''
#define NDIM {ndim}
#define CENTERED {centered}
typedef {cicType} BASIC_TYPE;

'''

CIC_KERNEL='''///CL///
#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable

__kernel void init_pcell(__global int *p_cell, const int value)
{
    int i = get_global_id(0);
    p_cell[i] = value;
}

__kernel void build_indices(__global const BASIC_TYPE *pos,
                            __global int *part_mesh, __global int *part_list, const int N, const BASIC_TYPE delta, const BASIC_TYPE shift_pos)
{
    int i_part = get_global_id(0);
    long shifter = 1;
    long idx = 0;
    int d;

    for (d = 0; d < NDIM; d++) {
        BASIC_TYPE x;
        
        if (CENTERED)
          x = pos[i_part*NDIM + d] - shift_pos;
        else
          x = pos[i_part*NDIM + d];

        int m = (int)floor(x*delta) %% N;
        
        idx += shifter * m;        
        shifter *= N;
    }
    
    // Head of the list
    int initial_elt = atom_xchg(&part_mesh[idx], i_part);
    if (initial_elt == -1) {
        return;
    }
    // Point the next pointer of old_end to i_part
    part_list[i_part] = initial_elt;
}

__kernel void reverse_list(__global int *part_mesh, __global int *part_list)
{
    int mid = get_global_id(0);
   
    int current_part = part_mesh[mid];
    if (current_part >= 0) {
        int next_part = part_list[current_part];
        part_list[current_part] = -1;
        while (next_part != -1) {
          int p = part_list[next_part];
          part_list[next_part] = current_part;
          current_part = next_part;
          next_part = p;
        }
        part_mesh[mid] = current_part;
    }
}

__kernel void dance(__global const BASIC_TYPE *pos,
                    __global BASIC_TYPE *density,
                    __global int *part_mesh, __global int *part_list, const int N, const BASIC_TYPE delta, const BASIC_TYPE shift_pos)
{
    int m[NDIM];
    int shifter = 1;    
    int i;
    int first, i_part;
    int idx = 0;
    
    for (i = 0; i < NDIM; i++) {
        m[i] = get_global_id(i);
        idx += shifter * m[i];
        shifter *= N;
    }

    first = 1;    

//BEGIN LOOPER
%(looperFor)s
//END LOOPER        

            int idx_dance = 0; 
            BASIC_TYPE w = 0;
//LOOPER INDEX
            int r[NDIM] = { %(looperVariables)s };
//END LOOPER

            i_part = part_mesh[idx];
            while (i_part != -1) {
                BASIC_TYPE w0 = 1;
                
                for (int d = 0; d < NDIM; d++) {
                    BASIC_TYPE x;
                    BASIC_TYPE q;
                    BASIC_TYPE dx;

                    if (CENTERED)
                      x = pos[i_part*NDIM + d]*delta - shift_pos;
                    else
                      x = pos[i_part*NDIM + d]*delta;
                    q = floor(x);
                    dx = x - q;
                    
                    w0 *= (r[d] == 1) ? dx : ((BASIC_TYPE)1-dx);
                }

                i_part = part_list[i_part];
                w += w0;
            }
        
            shifter = 1;    
            for (i = 0; i < NDIM; i++) {
                idx_dance += shifter * ((m[i]+r[i])%%N);
                shifter *= N;
            }

            density[idx_dance] += w;
                
            // One dance done. Wait for everybody for the next iteration
            barrier(CLK_GLOBAL_MEM_FENCE);
%(looperForEnd)s
}
'''

class CIC_CL(object):

    def __init__(self, context, ndim=2, ktype=np.float32, centered=False):
        global CIC_PREKERNEL, CIC_KERNEL

        translator = {}
        if ktype == np.float32:
            translator['cicType'] = 'float'
            pragmas = ''
        elif ktype == np.float64:
            translator['cicType'] = 'double'
            pragmas = '#pragma OPENCL EXTENSION cl_khr_fp64 : enable\n'
        else:
            raise ValueError("Invalid ktype")
            
        # 2 dimensions
        translator['ndim'] = ndim
        translator['centered'] = '1' if centered else '0'
        looperVariables = ','.join(['id%d' % d for d in range(ndim)])
        looperFor = '\n'.join(['for (int id{dim}=0; id{dim} < 2; id{dim}++) {{'.format(dim=d) for d in range(ndim)])
        looperForEnd = '}' * ndim
        
        kern = pragmas + CIC_PREKERNEL.format(**translator) + (CIC_KERNEL % {'looperVariables': looperVariables, 'looperFor': looperFor, 'looperForEnd':looperForEnd})
        self.kern_code = kern
        self.ctx = context
        self.queue = cl.CommandQueue(context)#, properties=cl.OUT_OF_ORDER_EXEC_MODE_ENABLE)
        self.ktype = ktype
        self.ndim = ndim
        self.prog = cl.Program(self.ctx, kern).build()
        self.centered = centered

    def run(self, particles, Ng, L):
        assert particles.strides[1] == self.ktype().itemsize  # This is C-ordering
        assert particles.shape[1] == self.ndim

        print("Start again")

        ndim = self.ndim
        part_pos = cl_array.to_device(self.queue, particles)
        part_mesh = cl_array.empty(self.queue, (Ng,)*ndim, np.int32, order='C')
        density = cl_array.zeros(self.queue, (Ng,)*ndim, self.ktype, order='C')
        part_list = cl_array.empty(self.queue, (particles.shape[0],), np.int32, order='C')
        shift_pos = 0.5*L if self.centered else 0
        
        if True:
          delta = Ng/L
    
          with time_block("Init pcell array"):
              e = self.prog.init_pcell(self.queue, (Ng**ndim,), None, part_mesh.data, np.int32(-1))
              e.wait()
        
          with time_block("Init idx array"):
              e=self.prog.init_pcell(self.queue, (particles.shape[0],), None, part_list.data, np.int32(-1))
              e.wait()
        
          with time_block("Build indices"):
              self.prog.build_indices(self.queue, (particles.shape[0],), None,
                      part_pos.data, part_mesh.data, part_list.data, np.int32(Ng), self.ktype(delta), self.ktype(shift_pos))
        
        if True:
          with time_block("Reverse list"):
              lastevt = self.prog.reverse_list(self.queue, (Ng**ndim,), None, part_mesh.data, part_list.data)
        # We require pmax pass, particles are ordered according to part_idx

          with time_block("dance"):
              self.prog.dance(self.queue, (Ng,)*ndim, None, part_pos.data, density.data, part_mesh.data, part_list.data, np.int32(Ng), self.ktype(delta), self.ktype(shift_pos))
    
        self.queue.finish()
        del part_pos
        del part_mesh
        del part_list
        with time_block("download"):
          return density.get()


def cl_CIC_Density(particles, Ngrid, Lbox, context=None, periodic=True, centered=False):
  """
  cl_CIC_Density(particles (Nx3), Ngrid, Lbox, context=None, periodic=True, centered=False)
"""
  if context is None:
    context = cl.create_some_context()

  ktype = particles.dtype
  if ktype != np.float32 and ktype != np.float64:
    raise ValueError("particles may only be float32 or float64")

  if len(particles.shape) != 2 or particles.shape[1] != 3:
    raise ValueError("particles may only be a Nx3 array")

  cic = CIC_CL(context, ndim=3, centered=centered, ktype=ktype)

  return cic.run(particles, Ngrid, Lbox)



from .config import install_prefix
import subprocess
import os
try:
  from tempfile import TemporaryDirectory
except:
  from backports.tempfile import TemporaryDirectory

import h5py as h5
import numpy as np
import weakref

def smooth_particle_density(
           position,
           velocities=None,
           radius=1e6,
           boxsize=None, 
           resolution=128,
           center=None, tmpprefix=None ):
  """Use adaptive smoothing to produce density and momentum fields.
  The algorithm is originally described in [1].

  Parameters:
  position   : numpy array NxQ
               the particle positions
               if Q==3, only positions. Q==6 means full space phase
  velocities : Optional numpy array Nx3. 
               It is only optional if the above Q is 6.
  radius     : float
               Maximum radius to which we need to compute fields
  boxsize    : float
               Size of the box for the generated fields
  resolution : int
               Resolution of the output boxes
  center     : list of 3 floats
               Center of the new box. It depends on the convention 
               for particles. If those are between [0, L], then [0,0,0]
               is correct. If those are [-L/2,L/2] then you should set
               [L/2,L/2,L/2].
  tmpprefix  : string
               prefix of the temporary directory that will be used.
               It needs to have a lot of space available. By default
               '/tmp/ will be typically used.


  Returns
  -------
   dictionnary
   The dict has two entries: 'rho' for the density, and 'p' for the momenta.
   Once the dictionary is garbage collected all temporary files and directories
   will be cleared automatically.


  Raises
  ------
   ValueError
     if arguments are invalid


  .. [1] S. Colombi, M. Chodorowski, 
         "Cosmic velocity-gravity in redshift space", MNRAS, 2007, 375, 1
  """
  if len(position.shape) != 2:
     raise ValueError("Invalid position array shape")

  if velocities is None:
    if position.shape[1] != 6:
      raise ValueError("Position must be phase space if no velocities are given")


  if boxsize is None:
    raise ValueError("Need a boxsize")

  cx,cy,cz=center
  tmpdir = TemporaryDirectory(prefix=tmpprefix) 
  h5_file = os.path.join(tmpdir.name, 'particles.h5')
  with h5.File(h5_file, mode="w") as f:
    data = f.create_dataset('particles', shape=(position.shape[0],7), dtype=np.float32)
    data[:,:3] = position[:,:3]
    if velocities is not None: 
      data[:,3:6] = velocities[:,:3]
    else:
      data[:,3:6] = position[:,3:]
    data[:,6] = 1


  ret =  \
     subprocess.run([
         os.path.join(install_prefix,'bin','simple3DFilter'),
         h5_file,
         str(radius),
         str(boxsize),
         str(resolution),
         str(cx), str(cy), str(cz)
     ], cwd=tmpdir.name)

  
  f0 = h5.File(os.path.join(tmpdir.name,'fields.h5'), mode="r")
  def cleanup_f0():
    f0.close()
    tmpdir.cleanup()

  class Dict(dict):
    pass

  t = Dict(rho=f0['density'], p=[f0['p0'], f0['p1'], f0['p2']])
  t._tmpdir_=tmpdir
  weakref.finalize(t, cleanup_f0)
  return t



import numpy as np
import sys
import h5py as h5
import cosmotool as ct

s = ct.loadRamsesAll(sys.argv[1], int(sys.argv[2]), doublePrecision=True, loadVelocity=True)


q = [p for p in s.getPositions()]
q += [p for p in s.getVelocities()]
q += [np.ones(q[0].size,dtype=q[0].dtype)]
q = np.array(q)

with h5.File("particles.h5", mode="w") as f:
  f.create_dataset("particles", data=q.transpose())



from cpython cimport bool
from cython cimport view
from cython.parallel import prange, parallel
from libc.math cimport sin, cos, abs, floor, round, sqrt
import numpy as np
cimport numpy as npx
cimport cython

__all__=["fast_interp"]


@cython.boundscheck(False)
@cython.cdivision(True)
def fast_interp(xmin0, dx0, A0, y0, out0, beyond_val=np.nan):
  cdef double rq, q
  cdef int iq
  cdef long i, Asize, ysize
  cdef npx.float64_t xmin, dx
  cdef npx.float64_t[:] out
  cdef npx.float64_t[:] A, y
  cdef npx.float64_t beyond

  beyond=beyond_val
  xmin = xmin0
  dx = dx0
  A = A0
  y = y0
  ysize = y.size
  out = out0
  Asize = A.size

  if out.size != ysize:
    raise ValueError("out and y must have the same size")

  with nogil:
    for i in prange(ysize):
      q = (y[i] - xmin) / dx
      iq = int(floor(q))
      rq = (q-iq)
      if iq+1 >= Asize or iq < 0:
        out[i] = beyond
      else:
        out[i] = rq * A[iq+1] + (1-rq)*A[iq]



import os
import h5py as h5
import numpy as np
import cosmotool as ct
import icgen as bic
import icgen.cosmogrowth as cg
import sys
import argparse

#ADAPT_SMOOTH="/home/bergeron1NS/lavaux/Software/cosmotool/build/sample/simple3DFilter"
ADAPT_SMOOTH="/home/guilhem/PROJECTS/cosmotool/build/sample/simple3DFilter"
cosmo={'omega_M_0':0.3175, 'h':0.6711}
cosmo['omega_lambda_0']=1-cosmo['omega_M_0']
cosmo['omega_k_0'] = 0
cosmo['omega_B_0']=0.049
cosmo['SIGMA8']=0.8344
cosmo['ns']=0.9624
N0=256

doSimulation=True

astart=1.

parser=argparse.ArgumentParser(description="Generate CIC density from 2LPT")
parser.add_argument('--start', type=int, required=True)
parser.add_argument('--end', type=int, required=True)
parser.add_argument('--step', type=int, required=True)
parser.add_argument('--base', type=str, required=True)
parser.add_argument('--N', type=int, default=256)
parser.add_argument('--output', type=str, default="fields_%d.h5")
parser.add_argument('--supersample', type=int, default=1)
args = parser.parse_args()



for i in [4629]:#xrange(args.start, args.end, args.step):
  print i

  pos,vel,density,N,L,_,_ = bic.run_generation("%s/initial_density_%d.dat" % (args.base,i), 0.001, astart, 
                                           cosmo, supersample=args.supersample, do_lpt2=True)

  q = pos + vel + [np.ones(vel[0].shape[0])]

  with h5.File("particles.h5", mode="w") as f:
    f.create_dataset("particles", data=np.array(q).transpose())
  
  os.system(ADAPT_SMOOTH + " %s %lg %lg %d %lg %lg %lg" % ("particles.h5", 3000000, L, args.N, 0, 0, 0))
  os.rename("fields.h5", args.output % i)



import pyfftw
import multiprocessing
import numpy as np
import numexpr as ne

class CubeFT(object):
  def __init__(self, L, N, max_cpu=-1, width=32):
  
      if width==32:
        fourier_type='complex64'
        real_type='float32'
      elif width==64:
        fourier_type='complex128'
        real_type='float64'
      else:
        raise ValueError("Invalid bitwidth (must be 32 or 64)")

      self.N = N
      self.align = pyfftw.simd_alignment
      self.L = float(L)
      self.max_cpu = multiprocessing.cpu_count() if max_cpu < 0 else max_cpu
      self._dhat = pyfftw.n_byte_align_empty((self.N,self.N,self.N//2+1), self.align, dtype=fourier_type)
      self._density = pyfftw.n_byte_align_empty((self.N,self.N,self.N), self.align, dtype=real_type)
      self._irfft = pyfftw.FFTW(self._dhat, self._density, axes=(0,1,2), direction='FFTW_BACKWARD', threads=self.max_cpu)#, normalize_idft=False)
      self._rfft = pyfftw.FFTW(self._density, self._dhat, axes=(0,1,2), threads=self.max_cpu) #, normalize_idft=False)

  def rfft(self):
      return ne.evaluate('c*a', out=self._dhat, local_dict={'c':self._rfft(normalise_idft=False),'a':(self.L/self.N)**3}, casting='unsafe')
      
  def irfft(self):
      return ne.evaluate('c*a', out=self._density, local_dict={'c':self._irfft(normalise_idft=False),'a':(1/self.L)**3}, casting='unsafe')
   
  def get_dhat(self):
      return self._dhat
  def set_dhat(self, in_dhat):
      self._dhat[:] = in_dhat 
  dhat = property(get_dhat, set_dhat, None)
      
  def get_density(self):
      return self._density      
  def set_density(self, d):
      self._density[:] = d 
  density = property(get_density, set_density, None)
      




#+
# This is CosmoTool (./build_tools/gather_sources.py) -- Copyright (C) Guilhem Lavaux (2007-2014)
#
# guilhem.lavaux@gmail.com
#
# This software is a computer program whose purpose is to provide a toolbox for cosmological
# data analysis (e.g. filters, generalized Fourier transforms, power spectra, ...)
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#+
import shutil
import tempfile
import re
from git import Repo,Tree,Blob

def apply_license(license, relimit, filename):
  header = re.sub(r'@FILENAME@', filename, license)

  f = file(filename)
  lines = f.read()
  f.close()

  lines = re.sub(relimit, '', lines)
  lines = header + lines

  with tempfile.NamedTemporaryFile(delete=False) as tmp_sources:
    tmp_sources.write(lines)

  shutil.move(tmp_sources.name, filename)


def apply_python_license(filename):
  license="""#+
# This is CosmoTool (@FILENAME@) -- Copyright (C) Guilhem Lavaux (2007-2014)
#
# guilhem.lavaux@gmail.com
#
# This software is a computer program whose purpose is to provide a toolbox for cosmological
# data analysis (e.g. filters, generalized Fourier transforms, power spectra, ...)
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#+
"""
   
  print("Shell/Python file: %s" % filename)
  relimit=r'^#\+\n(#.*\n)*#\+\n'
  apply_license(license, relimit, filename)


def apply_cpp_license(filename):
  license="""/*+
This is CosmoTool (@FILENAME@) -- Copyright (C) Guilhem Lavaux (2007-2014)

guilhem.lavaux@gmail.com

This software is a computer program whose purpose is to provide a toolbox for cosmological
data analysis (e.g. filters, generalized Fourier transforms, power spectra, ...)

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.
+*/
"""
  relimit = r'^(?s)/\*\+.*\+\*/\n'
  print("C++ file: %s" % filename)
  apply_license(license, relimit, filename)
  

def analyze_tree(prefix, t):
  for entry in t:
    ename = entry.name
    if ename == 'external' or ename == 'zobov':
      continue
    if type(entry) == Submodule:
      continue
    elif type(entry) == Tree:
      analyze_tree(prefix + "/" + ename, entry)
    elif type(entry) == Blob:
      if ename == './src/hdf5_flash.h' or ename == './src/h5_readFlash.cpp' or ename == './src/h5_readFlash.hpp':
        continue

      if re.match(".*\.(sh|py|pyx)$",ename) != None:
        fname=prefix+"/"+ename
        apply_python_license(fname)
      if re.match('.*\.(cpp|hpp|h)$', ename) != None:
        fname=prefix+"/"+ename
        apply_cpp_license(fname)


if __name__=="__main__":
  repo = Repo(".")
  assert repo.bare == False
  t = repo.tree()
  analyze_tree(".", t)



import stat
import os
import sys
import shutil
from distutils.command.install_data import install_data
import pathlib
from setuptools import find_packages, setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.install_lib import install_lib
from setuptools.command.install_scripts import install_scripts
import struct

BITS = struct.calcsize("P") * 8
PACKAGE_NAME = "cosmotool"

class CMakeExtension(Extension):
    """
    An extension to run the cmake build

    This simply overrides the base extension class so that setuptools
    doesn't try to build your sources for you
    """

    def __init__(self, name, sources=["python/dummy.c"]):

        super().__init__(name = name, sources = sources)

        self.SOURCE_DIR = str(pathlib.Path().absolute())

class InstallCMakeLibsData(install_data):
    """
    Just a wrapper to get the install data into the egg-info

    Listing the installed files in the egg-info guarantees that
    all of the package files will be uninstalled when the user
    uninstalls your package through pip
    """

    def run(self):
        """
        Outfiles are the libraries that were built using cmake
        """

        # There seems to be no other way to do this; I tried listing the
        # libraries during the execution of the InstallCMakeLibs.run() but
        # setuptools never tracked them, seems like setuptools wants to
        # track the libraries through package data more than anything...
        # help would be appriciated

        self.outfiles = self.distribution.data_files

class InstallCMakeLibs(install_lib):
    """
    Get the libraries from the parent distribution, use those as the outfiles

    Skip building anything; everything is already built, forward libraries to
    the installation step
    """

    def run(self):
        """
        Copy libraries from the bin directory and place them as appropriate
        """

        self.announce("Moving library files", level=3)
#        print(self.build_temp)
#
#        self.distribution.bin_dir = CosmoTool_extension.bin_dir
#
#        # We have already built the libraries in the previous build_ext step
#
#        self.skip_build = True
#
#        # Depending on the files that are generated from your cmake
#        # build chain, you may need to change the below code, such that
#        # your files are moved to the appropriate location when the installation
#        # is run
#
#        libs = [os.path.join(bin_dir, _lib) for _lib in 
#                os.listdir(bin_dir) if 
#                os.path.isfile(os.path.join(bin_dir, _lib)) and 
#                os.path.splitext(_lib)[1] in [".dll", ".so"]
#                and not (_lib.startswith("python") or _lib.startswith(PACKAGE_NAME))]
#
#        for lib in libs:
#
#            shutil.move(lib, os.path.join(self.build_dir,
#                                          os.path.basename(lib)))
#
#        # Mark the libs for installation, adding them to 
#        # distribution.data_files seems to ensure that setuptools' record 
#        # writer appends them to installed-files.txt in the package's egg-info
#        #
#        # Also tried adding the libraries to the distribution.libraries list, 
#        # but that never seemed to add them to the installed-files.txt in the 
#        # egg-info, and the online recommendation seems to be adding libraries 
#        # into eager_resources in the call to setup(), which I think puts them 
#        # in data_files anyways. 
#        # 
#        # What is the best way?
#
#        # These are the additional installation files that should be
#        # included in the package, but are resultant of the cmake build
#        # step; depending on the files that are generated from your cmake
#        # build chain, you may need to modify the below code
#
#        self.distribution.data_files = [os.path.join(self.install_dir, 
#                                                     os.path.basename(lib))
#                                        for lib in libs]
#        print(self.distribution.data_files)
#
#        # Must be forced to run after adding the libs to data_files

        self.distribution.run_command("install_data")

        super().run()

class BuildCMakeExt(build_ext):
    """
    Builds using cmake instead of the python setuptools implicit build
    """

    def run(self):
        """
        Perform build_cmake before doing the 'normal' stuff
        """

        for extension in self.extensions:

            if extension.name == 'cosmotool':
                self.package = 'cosmotool'

                self.build_cmake(extension)

        super().run()

    def build_cmake(self, extension: Extension):
        """
        The steps required to build the extension
        """

        self.announce("Preparing the build environment", level=3)

        package_dir = os.path.abspath(os.path.join(self.build_lib, 'cosmotool'))


        extension.build_dir = pathlib.Path(self.build_temp)
        extension.bin_dir = str(pathlib.Path(os.path.join(extension.build_dir, 'private_install')).absolute())
        SOURCE_DIR = extension.SOURCE_DIR
        build_dir = extension.build_dir

        extension_path = pathlib.Path(self.get_ext_fullpath(extension.name))

        os.makedirs(build_dir, exist_ok=True)
        os.makedirs(extension_path.parent.absolute(), exist_ok=True)

        cython_code = os.path.join(str(build_dir.absolute()),'mycython')
        with open(cython_code, mode="wt") as ff:
          ff.write("#!/bin/sh\n"
                   f"{sys.executable} -c 'from Cython.Compiler.Main import setuptools_main; setuptools_main()' $@")
        os.chmod(cython_code, stat.S_IXUSR|stat.S_IWUSR|stat.S_IRUSR|stat.S_IRGRP)

        # Now that the necessary directories are created, build

        self.announce("Configuring cmake project", level=3)

        # Change your cmake arguments below as necessary
        # Below is just an example set of arguments for building Blender as a Python module

        compilers=[]
        if "CC" in os.environ:
           compilers.append('-DCMAKE_C_COMPILER=' + os.environ["CC"])
        if "CXX" in os.environ:
           compilers.append("-DCMAKE_CXX_COMPILER=" + os.environ["CXX"])

        self.spawn(['cmake', '-H'+SOURCE_DIR, '-B'+self.build_temp,
                    '-DENABLE_OPENMP=ON','-DINTERNAL_BOOST=ON','-DINTERNAL_EIGEN=ON',
                    '-DINTERNAL_HDF5=ON','-DINTERNAL_NETCDF=ON',
                    '-DBUILD_PYTHON=ON', '-DINSTALL_PYTHON_LOCAL=OFF',
                    '-DCOSMOTOOL_PYTHON_PACKAGING=ON',
                    f"-DCYTHON={cython_code}",
                    f"-DPYTHON_SITE_PACKAGES={build_dir.absolute()}/private_install",
                    f"-DPYTHON_EXECUTABLE={sys.executable}"] + compilers)


        self.announce("Building binaries", level=3)

        self.spawn(["cmake", "--build", self.build_temp, "--target", "install",
                      "--config", "Release"])

        # Build finished, now copy the files into the copy directory
        # The copy directory is the parent directory of the extension (.pyd)

        self.announce("Moving built python module", level=3)

        bin_dir = extension.bin_dir
        self.distribution.bin_dir = bin_dir

        pyd_path=[]
        top_level_len = len(pathlib.Path(bin_dir).parts)
        for root,_,_pyds in os.walk(bin_dir):
          for _pyd in _pyds:
            print(_pyd)
            _pyd=os.path.join(root,_pyd)
            if os.path.isfile(_pyd):
               _pyd_top = pathlib.Path(_pyd).parts[top_level_len:]
               if  _pyd_top[0].startswith(PACKAGE_NAME):
                   if os.path.splitext(_pyd)[1] in [".pyd", ".so"] or _pyd_top[-1] == 'config.py':
                     pyd_path.append((_pyd_top,_pyd))
               

        for top,p in pyd_path:
           _,n = os.path.split(p)
           n,e = os.path.splitext(n)
           if n != "_cosmo_bispectrum" and n != 'config':
             new_p = pathlib.Path(self.get_ext_fullpath(n))
           else:
             print(f"package_dir is {package_dir}")
             new_p = pathlib.Path(os.path.join(package_dir,n+e))
           self.announce(f"Moving {p} to {new_p}", level=3)
           shutil.move(p, new_p)

CosmoTool_extension = CMakeExtension(name="cosmotool")

setup(name='cosmotool',
      version='1.2.0',
      packages=["cosmotool"],
      package_dir={'cosmotool': 'python/cosmotool'},
      install_requires=['numpy','cffi','numexpr','pyfftw','h5py'],
      setup_requires=['cython','cffi','numpy','numexpr'],
      ext_modules=[CosmoTool_extension],
      description='A small cosmotool box of useful functions',
      long_description=open("./README.md", 'r').read(),
      long_description_content_type="text/markdown",
      keywords="cosmology, interpolation, cmake, extension",
      classifiers=["Intended Audience :: Developers",
                   "License :: OSI Approved :: "
                   "GNU Lesser General Public License v3 (LGPLv3)",
                   "Natural Language :: English",
                   "Programming Language :: C",
                   "Programming Language :: C++",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: Implementation :: CPython"],
      license='CeCILL-v2',
      cmdclass={
          'build_ext': BuildCMakeExt,
          'install_data': InstallCMakeLibsData,
          'install_lib': InstallCMakeLibs,
          },
      python_requires='>=3.6'
)



import numexpr as ne
import multiprocessing
import pyfftw
import weakref
import numpy as np
import cosmolopy as cpy
import cosmotool as ct

class CubeFT(object):
  def __init__(self, L, N, max_cpu=-1):
  
      self.N = N
      self.align = pyfftw.simd_alignment
      self.L = L
      self.max_cpu = multiprocessing.cpu_count() if max_cpu < 0 else max_cpu
      self._dhat = pyfftw.n_byte_align_empty((self.N,self.N,self.N/2+1), self.align, dtype='complex64')
      self._density = pyfftw.n_byte_align_empty((self.N,self.N,self.N), self.align, dtype='float32')
      self._irfft = pyfftw.FFTW(self._dhat, self._density, axes=(0,1,2), direction='FFTW_BACKWARD', threads=self.max_cpu, normalize_idft=False)
      self._rfft = pyfftw.FFTW(self._density, self._dhat, axes=(0,1,2), threads=self.max_cpu, normalize_idft=False)

  def rfft(self):
      return ne.evaluate('c*a', local_dict={'c':self._rfft(normalise_idft=False),'a':(self.L/self.N)**3})
      
  def irfft(self):
      return ne.evaluate('c*a', local_dict={'c':self._irfft(normalise_idft=False),'a':(1/self.L)**3})
   
  def get_dhat(self):
      return self._dhat
  def set_dhat(self, in_dhat):
      self._dhat[:] = in_dhat 
  dhat = property(get_dhat, set_dhat, None)
      
  def get_density(self):
      return self._density      
  def set_density(self, d):
      self._density[:] = d 
  density = property(get_density, set_density, None)
      

class CosmoGrowth(object):

  def __init__(self, **cosmo):
    self.cosmo = cosmo

  def D(self, a):
    return cpy.perturbation.fgrowth(1/a-1, self.cosmo['omega_M_0'], unnormed=True)

  def compute_E(self, a):
    om = self.cosmo['omega_M_0']
    ol = self.cosmo['omega_lambda_0']
    ok = self.cosmo['omega_k_0']

    E = np.sqrt(om/a**3 + ol + ok/a**2)
  
    H2 = -3*om/a**4  - 2*ok/a**3

    Eprime = 0.5*H2/E

    return E,Eprime

  def Ddot(self, a):
    E,Eprime = self.compute_E(a)
    D = self.D(a)
    Ddot_D = Eprime/E + 2.5 * self.cosmo['omega_M_0']/(a**3*E**2*D)
    Ddot_D *= a
    return Ddot_D

  def compute_velmul(self, a):
    E,_ = self.compute_E(a)
    velmul = self.Ddot(a)
    velmul *= 100 * a * E
    return velmul





class LagrangianPerturbation(object):

    def __init__(self,density,L, fourier=False, supersample=1, max_cpu=-1):
    
        self.L = L
        self.N = density.shape[0]
        
        self.max_cpu = max_cpu
        self.cube = CubeFT(self.L, self.N, max_cpu=max_cpu)
        
        if not fourier:
          self.cube.density = density
          self.dhat = self.cube.rfft().copy()
        else:
          self.dhat = density.copy()

        if supersample > 1:
          self.upgrade_sampling(supersample)
        self.ik = np.fft.fftfreq(self.N, d=L/self.N)*2*np.pi
        self._kx = self.ik[:,None,None]
        self._ky = self.ik[None,:,None]
        self._kz = self.ik[None,None,:(self.N/2+1)]
        self.cache = {}#weakref.WeakValueDictionary()

    @ct.timeit_quiet
    def upgrade_sampling(self, supersample):
        N2 = self.N * supersample
        N = self.N
        dhat_new = np.zeros((N2, N2, N2/2+1), dtype=np.complex128)

        hN = N/2
        dhat_new[:hN, :hN, :hN+1] = self.dhat[:hN, :hN, :]
        dhat_new[:hN, (N2-hN):N2, :hN+1] = self.dhat[:hN, hN:, :]
        dhat_new[(N2-hN):N2, (N2-hN):N2, :hN+1] = self.dhat[hN:, hN:, :]
        dhat_new[(N2-hN):N2, :hN, :hN+1] = self.dhat[hN:, :hN, :]

        self.dhat = dhat_new
        self.N = N2
        self.cube = CubeFT(self.L, self.N, max_cpu=self.max_cpu)
        
    @ct.timeit_quiet
    def _gradient(self, phi, direction):
        if direction == 'all':
          dirs = [0,1,2]
          copy = True
        else:
          dirs = [direction]
          copy = False
        ret=[]
        for dir in dirs:
          ne.evaluate('phi_hat * i * kv / (kx**2 + ky**2 + kz**2)', out=self.cube.dhat,
              local_dict={'i':-1j, 'phi_hat':phi, 'kv':self._kdir(dir),
                          'kx':self._kx, 'ky':self._ky, 'kz':self._kz},casting='unsafe')
#        self.cube.dhat = self._kdir(direction)*1j*phi 
          self.cube.dhat[0,0,0] = 0
          x = self.cube.irfft()
          ret.append(x.copy() if copy else x)
        return ret[0] if len(ret)==1 else ret
            
    @ct.timeit_quiet
    def lpt1(self, direction=0):
        return self._gradient(self.dhat, direction)
        
    def new_shape(self,direction, q=3, half=False):
        N0 = (self.N/2+1) if half else self.N
        return ((1,)*direction) + (N0,) + ((1,)*(q-1-direction))

    def _kdir(self, direction, q=3):
        if direction != q-1:
            return self.ik.reshape(self.new_shape(direction, q=q))
        else:
            return self.ik[:self.N/2+1].reshape(self.new_shape(direction, q=q, half=True))

    def _get_k2(self, q=3):
        if 'k2' in self.cache:
            return self.cache['k2']
            
        k2 = self._kdir(0, q=q)**2
        for d in xrange(1,q):
            k2 = k2 + self._kdir(d, q=q)**2
            
        self.cache['k2'] = k2
        return k2
        
    def _do_irfft(self, array, copy=True):
        if copy:
          self.cube.dhat = array
        return self.cube.irfft()
        
    def _do_rfft(self, array, copy=True):
        if copy:
          self.cube.density = array
        return self.cube.rfft()

    @ct.timeit_quiet
    def lpt2(self, direction=0):
#        k2 = self._get_k2()
#        k2[0,0,0] = 1

        inv_k2 = ne.evaluate('1/(kx**2+ky**2+kz**2)', {'kx':self._kdir(0),'ky':self._kdir(1),'kz':self._kdir(2)})
        inv_k2[0,0,0]=0
        potgen0 = lambda i: ne.evaluate('kdir**2*d*ik2',out=self.cube.dhat,local_dict={'kdir':self._kdir(i),'d':self.dhat,'ik2':inv_k2}, casting='unsafe' )
        potgen = lambda i,j: ne.evaluate('kdir0*kdir1*d*ik2',out=self.cube.dhat,local_dict={'kdir0':self._kdir(i),'kdir1':self._kdir(j),'d':self.dhat,'ik2':inv_k2}, casting='unsafe' )

        if 'lpt2_potential' not in self.cache:
            print("Rebuilding potential...")
            div_phi2 = np.zeros((self.N,self.N,self.N), dtype=np.float64)
            for j in xrange(3):
                q = self._do_irfft( potgen0(j) ).copy()
                for i in xrange(j+1, 3):
                    with ct.time_block("LPT2 elemental (%d,%d)" %(i,j)):
                      ne.evaluate('div + q * pot', out=div_phi2, 
                            local_dict={'div':div_phi2, 'q':q,'pot':self._do_irfft( potgen0(i), copy=False ) } 
                            )
                      ne.evaluate('div - pot**2',out=div_phi2,
                            local_dict={'div':div_phi2,'pot':self._do_irfft(potgen(i,j), copy=False) }
                            )

            phi2_hat = self._do_rfft(div_phi2)
            #self.cache['lpt2_potential'] = phi2_hat
            del div_phi2
        else:
            phi2_hat = self.cache['lpt2_potential']
            
        return self._gradient(phi2_hat, direction)



import numexpr as ne
import numpy as np


def cicParticles(particles, L, N):

    if type(N) not in [int,int]:
      raise TypeError("N must be a numeric type")

    def shifted(i, t):
      a = np.empty(i[0].size, dtype=np.int64)
      return ne.evaluate('(i2+t2)%N + N*((i1+t1)%N + N*((i0+t0)%N) )', local_dict={'i2':i[2], 't2':t[2], 'i1':i[1], 't1':t[1], 'i0':i[0], 't0':t[0], 'N':N}, out=a)


    i =[]
    r = []
    for d in range(3):
        q = ne.evaluate('(p%L)*N/L', local_dict={'p':particles[d], 'L':L, 'N':N })
        o = np.empty(q.size, dtype=np.int64)
        o[:] = np.floor(q)
        i.append(o)
        r.append(ne.evaluate('q-o'))

    D = {'a':r[0],'b':r[1],'c':r[2]}
    N3 = N*N*N

    def accum(density, ss, op):
      d0 = np.bincount(shifted(i, ss), weights=ne.evaluate(op, local_dict=D), minlength=N3)
      ne.evaluate('d + d0', local_dict={'d':density, 'd0':d0}, out=density)

    density = np.empty(N3, dtype=np.float64)

    accum(density, (1,1,1), 'a     *    b  *    c ')
    accum(density, (1,1,0), 'a     *    b  * (1-c)')
    accum(density, (1,0,1), 'a     * (1-b) *    c ')
    accum(density, (1,0,0), 'a     * (1-b) * (1-c)')
    accum(density, (0,1,1), '(1-a) *    b  *    c ')
    accum(density, (0,1,0), '(1-a) *    b  * (1-c)')
    accum(density, (0,0,1), '(1-a) * (1-b) *    c ')
    accum(density, (0,0,0), '(1-a) * (1-b) * (1-c)')

    return density.reshape((N,N,N))
